from unittest.mock import ANY, Mock, patch

import pytest
from click.testing import CliRunner
from pygitguardian.models import Detail

from ggshield.__main__ import cli
from ggshield.core.config.user_config import SecretConfig
from ggshield.core.errors import ExitCode
from ggshield.core.scan import StringScannable
from ggshield.utils.git_shell import EMPTY_SHA, Filemode
from ggshield.utils.os import cd
from ggshield.verticals.secret import Result, Results, SecretScanCollection
from tests.repository import Repository, create_pre_receive_repo
from tests.unit.conftest import (
    _SIMPLE_SECRET_PATCH,
    _SIMPLE_SECRET_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_TOKEN,
    assert_invoke_exited_with,
    assert_invoke_ok,
)


def contains_secret(line: str, secret: str) -> bool:
    """Returns True if `line` contains an obfuscated version of `secret`"""
    return f'"{secret[:6]}' in line and f'{secret[-6:]}"' in line


def mock_multiprocessing_process(mock: Mock):
    """Mock to execute the target and return the mock object"""

    def mock_constructor(target, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        try:
            target(*args, **kwargs)
        except SystemExit as exit_exc:
            mock.exitcode = exit_exc.code

        mock.is_alive.return_value = False
        return mock

    mock.is_alive.return_value = True
    mock.exitcode = None
    return mock_constructor


class TestPreReceive:
    @pytest.fixture(autouse=True)
    def mock_multiprocessing(self):
        """
        multiprocessing.Process is mocked to make everything run on the main process
        to permit mocking of scan_commit_range
        """
        with patch(
            "ggshield.cmd.secret.scan.prereceive.multiprocessing"
        ) as multiprocessing_mock:
            multiprocessing_mock.Process.side_effect = mock_multiprocessing_process(
                multiprocessing_mock.Process
            )
            yield

    @patch("ggshield.cmd.secret.scan.prereceive.scan_commit_range")
    def test_stdin_input(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN 3 commits through stdin input
        WHEN the command is run
        THEN it should pass onto scan and return 0
        """
        scan_commit_range_mock.return_value = ExitCode.SUCCESS

        repo = create_pre_receive_repo(tmp_path)
        old_sha = repo.get_top_sha()
        shas = [repo.create_commit() for _ in range(3)]
        with cd(repo.path):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{old_sha} {shas[-1]} origin/main\n",
            )
        assert_invoke_ok(result)
        scan_commit_range_mock.assert_called_once()
        assert "Commits to scan: 3" in result.output

    @patch("ggshield.cmd.secret.scan.prereceive.scan_commit_range")
    def test_stdin_input_secret(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN 3 commits through stdin input
        WHEN the command is run and there are secrets
        THEN it should return a special remediation message
        """
        scan_commit_range_mock.return_value = ExitCode.SCAN_FOUND_PROBLEMS

        repo = create_pre_receive_repo(tmp_path)
        old_sha = repo.get_top_sha()
        shas = [repo.create_commit() for _ in range(3)]
        with cd(repo.path):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{old_sha} {shas[-1]} origin/main\n",
            )
        assert_invoke_exited_with(result, ExitCode.SCAN_FOUND_PROBLEMS)
        scan_commit_range_mock.assert_called_once()
        assert (
            """> How to remediate

  A pre-receive hook set server side prevented you from pushing secrets.

  Since the secret was detected during the push BUT after the commit, you need to:
  1. rewrite the git history making sure to replace the secret with its reference (e.g. environment variable).
  2. push again.

  To prevent having to rewrite git history in the future, setup ggshield as a pre-commit hook:
    https://docs.gitguardian.com/ggshield-docs/integrations/git-hooks/pre-commit

> [To apply with caution] If you want to bypass ggshield (false positive or other reason), run:

    git push -o breakglass"""
            in result.output
        )

    @patch("ggshield.cmd.secret.scan.prereceive.scan_commit_range")
    def test_stdin_breakglass_2ndoption(
        self,
        scan_commit_range_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN commits sent through stdin but breakglass active
        WHEN the command is run
        THEN it should return 0
        AND no commits should be scanned
        """
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "pre-receive"],
            input="bbbb aaaa origin/main\n",
            env={
                "GIT_PUSH_OPTION_COUNT": "2",
                "GIT_PUSH_OPTION_0": "unrelated",
                "GIT_PUSH_OPTION_1": "breakglass",
            },
        )
        assert_invoke_ok(result)
        scan_commit_range_mock.assert_not_called()
        assert (
            "SKIP: breakglass detected. Skipping GitGuardian pre-receive hook.\n"
            in result.output
        )

    @patch("ggshield.verticals.secret.repo.check_client_api_key")
    @patch("ggshield.verticals.secret.repo.scan_commits_content")
    def test_stdin_supports_gitlab_web_ui(
        self,
        scan_commits_content_mock: Mock,
        check_client_api_key_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN 1 webpush commit
        WHEN the command is run and there are secrets
        THEN it should return a special remediation message
        AND the GL-HOOK-ERR line should be there
        AND it should contain an obfuscated version of the secret
        """
        repo = create_pre_receive_repo(tmp_path)
        old_sha = repo.get_top_sha()
        secret_file = repo.path / "server.conf"
        secret_file.write_text(f"github_token = {_SIMPLE_SECRET_TOKEN}\n")
        repo.add(secret_file)
        secret_sha = repo.create_commit()

        check_client_api_key_mock.return_value = None

        # This test cannot mock scan_commit_range(): if it did that we would not get
        # the GitLab-specific output because output_handler.process_scan() would not be
        # called.
        scan_commits_content_mock.return_value = SecretScanCollection(
            id="some_id",
            type="commit-ranges",
            scans=[
                SecretScanCollection(
                    secret_sha,
                    type="commit",
                    results=Results(
                        results=[
                            Result.from_scan_result(
                                file=StringScannable(
                                    content=_SIMPLE_SECRET_PATCH,
                                    url="server.conf",
                                    filemode=Filemode.MODIFY,
                                ),
                                scan_result=_SIMPLE_SECRET_PATCH_SCAN_RESULT,
                                secret_config=SecretConfig(),
                            )
                        ],
                        errors=[],
                    ),
                )
            ],
        )

        with cd(repo.path):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{old_sha} {secret_sha} origin/main\n",
                env={
                    "GL_PROTOCOL": "web",
                },
            )
        assert_invoke_exited_with(result, ExitCode.SCAN_FOUND_PROBLEMS)
        scan_commits_content_mock.assert_called_once()
        web_ui_lines = [
            x for x in result.output.splitlines() if x.startswith("GL-HOOK-ERR: ")
        ]
        assert web_ui_lines
        assert any(contains_secret(x, _SIMPLE_SECRET_TOKEN) for x in web_ui_lines)

    def test_stdin_input_empty(
        self,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN an empty stdin input
        WHEN the command is run
        THEN it should raise an error and return 1
        """

        result = cli_fs_runner.invoke(
            cli, ["-v", "secret", "scan", "pre-receive"], input=""
        )
        assert_invoke_exited_with(result, ExitCode.UNEXPECTED_ERROR)
        assert "Error: Invalid input arguments: ''\n" in result.output

    @patch("ggshield.cmd.secret.scan.prereceive.scan_commit_range")
    def test_changing_max_commit_hooks(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a ref creation event
        WHEN the command is run with a changed env variable for max commit hooks
        THEN it should scan the last 2 commits
        """
        repo = create_pre_receive_repo(tmp_path)
        old_sha = repo.get_top_sha()
        shas = [repo.create_commit() for _ in range(3)]

        scan_commit_range_mock.return_value = ExitCode.SUCCESS

        with cd(repo.path):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{old_sha} {shas[-1]} main",
                env={"GITGUARDIAN_MAX_COMMITS_FOR_HOOK": "2"},
            )

        assert_invoke_ok(result)
        assert "Too many commits. Scanning last 2 commits" in result.output
        assert "Commits to scan: 2" in result.output
        scan_commit_range_mock.assert_called_once()

    @patch("ggshield.cmd.secret.scan.prereceive.scan_commit_range")
    def test_new_branch(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a repository
        AND commits received by a push of a new branch from another repository
        WHEN the pre-receive command is run on the commits from the push
        THEN it should scan only scan the commits of the new branch
        """
        repo = Repository.create(tmp_path)
        repo.create_commit("initial commit")

        # Detach from the current branch to simulate what happens when pre-receive
        # is called: the new commits are not in any branch yet.
        repo.git("checkout", "--detach")
        shas = [repo.create_commit() for _ in range(3)]

        scan_commit_range_mock.return_value = 0

        with cd(str(repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{EMPTY_SHA} {shas[-1]} refs/heads/topic",
            )

        assert_invoke_ok(result)
        scan_commit_range_mock.assert_called_once_with(
            client=ANY,
            cache=ANY,
            commit_list=shas,
            output_handler=ANY,
            exclusion_regexes=ANY,
            scan_context=ANY,
            secret_config=ANY,
        )

    @patch("ggshield.cmd.secret.scan.prereceive.scan_commit_range")
    def test_new_branch_without_commits(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a repository
        AND a new branch pushed from another repository, without any commit
        WHEN the pre-receive command is run on the commits from the push
        THEN it scans nothing
        """
        repo = Repository.create(tmp_path)
        sha = repo.create_commit("initial commit")
        branch_name = "topic"
        repo.create_branch(branch_name)

        with cd(str(repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{EMPTY_SHA} {sha} refs/heads/{branch_name}",
            )

        assert_invoke_ok(result)
        scan_commit_range_mock.assert_not_called()
        assert "Pushed branch does not contain any new commit" in result.output

    def test_stdin_input_deletion(
        self,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a deletion event
        WHEN the command is run
        THEN it should return 0 and indicate nothing to do
        """

        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "pre-receive"],
            input=f"{'a'*40} {EMPTY_SHA} main",
        )
        assert_invoke_ok(result)
        assert "Deletion event or nothing to scan.\n" in result.output

    @patch(
        "pygitguardian.client.GGClient.read_metadata",
        return_value=Detail("Service is unavailable", 503),
    )
    def test_server_unavailable(self, _: Mock, tmp_path, cli_fs_runner: CliRunner):
        """
        GIVEN a repo on which the command is ran
        WHEN the server is not responding (503)
        THEN it should return 0
        AND display an error message
        """
        # setting up repo to run the command
        repo = create_pre_receive_repo(tmp_path)
        old_sha = repo.get_top_sha()
        shas = [repo.create_commit() for _ in range(3)]
        with cd(repo.path):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-receive"],
                input=f"{old_sha} {shas[-1]} origin/main\n",
            )
        assert_invoke_ok(result)
        assert "GitGuardian server is not responding" in result.output



from typing import Any

import click

from ggshield.cmd.utils.common_options import add_common_options

from .login import login_cmd
from .logout import logout_cmd


@click.group(commands={"login": login_cmd, "logout": logout_cmd})
@add_common_options()
def auth_group(**kwargs: Any) -> None:
    """Commands to manage authentication."""



import json
import os
import shutil
import subprocess
from pathlib import Path
from string import Template

import jsonschema
import pytest

from ggshield.core.dirs import get_cache_dir
from tests.conftest import GG_VALID_TOKEN, skipwindows
from tests.functional.conftest import FUNCTESTS_DATA_PATH, requires_docker
from tests.functional.utils import recreate_censored_content, run_ggshield_scan


TEST_DOCKER_IMAGE = os.getenv("GGTEST_DOCKER_IMAGE", "ubuntu:20.04")


pytestmark = requires_docker()


@pytest.fixture()
def clear_layer_cache():
    shutil.rmtree(get_cache_dir() / "docker", ignore_errors=True)


def build_image(tmp_path: Path, name: str) -> None:
    image_path = tmp_path / "image"
    shutil.copytree(FUNCTESTS_DATA_PATH / name, image_path)

    variables = {
        "TEST_DOCKER_IMAGE": TEST_DOCKER_IMAGE,
        "GG_VALID_TOKEN": GG_VALID_TOKEN,
    }
    for path in image_path.rglob("*"):
        if not path.is_file():
            continue
        tmpl = Template(path.read_text())
        content = tmpl.substitute(variables)
        path.write_text(content)

    subprocess.run(["docker", "build", "-t", name, "."], cwd=image_path, check=True)


def test_scan_docker(clear_layer_cache) -> None:
    run_ggshield_scan("docker", TEST_DOCKER_IMAGE)


def test_scan_docker_json(clear_layer_cache, secret_json_schema) -> None:
    # GIVEN a scan of a docker image
    proc = run_ggshield_scan("docker", TEST_DOCKER_IMAGE, "--json")
    # THEN the output is a valid JSON, matching the secret schema
    dct = json.loads(proc.stdout)
    jsonschema.validate(dct, secret_json_schema)


@pytest.mark.parametrize(
    "image_name",
    (
        "docker-leaking-in-env",
        "docker-leaking-in-layer",
    ),
)
def test_scan_docker_find_secret(
    clear_layer_cache, tmp_path: Path, image_name: str
) -> None:
    build_image(tmp_path, image_name)
    proc = run_ggshield_scan("docker", image_name, expected_code=1, cwd=tmp_path)

    assert (
        recreate_censored_content(f"token={GG_VALID_TOKEN}", GG_VALID_TOKEN)
        in proc.stdout
    )


# Skip this test on Windows because on the CI the generated Docker image contains only
# one layer, so there is no caching.
# This is most likely related to the fact that the CI uses Windows Docker engine: the
# same test passes on a Windows machine using Linux Docker engine.
@skipwindows
def test_scan_docker_uses_cache(clear_layer_cache, tmp_path: Path) -> None:
    """
    GIVEN a docker image with a secret in the last layer
    WHEN scanned a second time
    THEN only the last layer is re-scanned
    """
    image_name = "docker-leaking-in-layer"
    scanning_snippet = "Scanning layer"
    skipping_snippet = "Skipping layer"

    build_image(tmp_path, image_name)
    proc = run_ggshield_scan("docker", image_name, expected_code=1, cwd=tmp_path)

    scanning_occurrences1 = proc.stderr.count(scanning_snippet)
    skipping_occurrences = proc.stderr.count(skipping_snippet)

    assert skipping_occurrences == 0

    proc = run_ggshield_scan("docker", image_name, expected_code=1, cwd=tmp_path)
    scanning_occurrences2 = proc.stderr.count(scanning_snippet)
    skipping_occurrences = proc.stderr.count(skipping_snippet)

    assert skipping_occurrences == 1
    assert scanning_occurrences2 == scanning_occurrences1 - 1



"""
This module centralizes GGShield error handling. For more details, have a look
at doc/dev/error-handling.md.
"""

import logging
import platform
import traceback
from enum import IntEnum
from typing import Any, Dict, List

import click
from marshmallow import ValidationError
from pygitguardian.models import Detail, TokenScope

from ggshield.core.text_utils import pluralize
from ggshield.utils.git_shell import GitError, InvalidGitRefError


logger = logging.getLogger(__name__)


class ExitCode(IntEnum):
    """
    Define constant exit codes based on their type
    """

    # Everything went well
    SUCCESS = 0
    # Scan was successful, and found problems (e.g. leaked secrets)
    SCAN_FOUND_PROBLEMS = 1
    # Error on the command-line, like a missing parameter
    USAGE_ERROR = 2
    # auth subcommand failed
    AUTHENTICATION_ERROR = 3
    # GitGuardian server is not responding
    GITGUARDIAN_SERVER_UNAVAILABLE = 4

    # Add new exit codes here.
    # If you add a new exit code, make sure you also add it to the documentation.

    # Catch all for other failures
    UNEXPECTED_ERROR = 128


class _ExitError(click.ClickException):
    """
    Base class for exceptions which must exit with an exit code as defined in ExitCode.

    This class is internal, inherit from it to create public exception classes.
    """

    def __init__(self, exit_code: ExitCode, message: str) -> None:
        super().__init__(message)
        self.exit_code = exit_code


class UnexpectedError(_ExitError):
    def __init__(self, message: str) -> None:
        super().__init__(ExitCode.UNEXPECTED_ERROR, message)


class ParseError(_ExitError):
    """
    Failed to load file
    """

    def __init__(self, message: str):
        super().__init__(ExitCode.UNEXPECTED_ERROR, message)


class MissingScopesError(_ExitError):
    """
    Token does not have the required scope
    """

    def __init__(self, token_scopes: List[TokenScope]):
        scopes_list = ", ".join([scope.value for scope in token_scopes])
        super().__init__(
            ExitCode.UNEXPECTED_ERROR,
            f"Token is missing the required {pluralize('scope', len(token_scopes))} {scopes_list} to perform this operation.",  # noqa
        )


class AuthError(_ExitError):
    """
    Base exception for Auth-related configuration error
    """

    def __init__(self, instance: str, message: str):
        super().__init__(ExitCode.AUTHENTICATION_ERROR, message)
        self.instance = instance


class QuotaLimitReachedError(_ExitError):
    def __init__(self):
        super().__init__(
            ExitCode.UNEXPECTED_ERROR,
            "Could not perform the requested action: no more API calls available.",
        )


class UnknownInstanceError(AuthError):
    """
    Raised when the requested instance does not exist
    """

    def __init__(self, instance: str):
        super().__init__(instance, f"Unknown instance: '{instance}'")


class AuthExpiredError(AuthError):
    """
    Raised when authentication has expired for the given instance
    """

    def __init__(self, instance: str):
        super().__init__(
            instance,
            f"Instance '{instance}' authentication expired, please authenticate again.",
        )


class MissingTokenError(AuthError):
    def __init__(self, instance: str):
        super().__init__(instance, f"No token is saved for this instance: '{instance}'")


class APIKeyCheckError(AuthError):
    """
    Raised when checking the API key fails
    """

    def __init__(self, instance: str, message: str):
        super().__init__(instance, message)


class ServiceUnavailableError(_ExitError):
    """
    Raised when the server is unavailable
    """

    def __init__(self, message: str):
        super().__init__(ExitCode.GITGUARDIAN_SERVER_UNAVAILABLE, message)


def format_validation_error(exc: ValidationError) -> str:
    """
    Take a Marshmallow ValidationError and turn it into a more user-friendly message
    """
    message_dct = exc.normalized_messages()
    lines = []

    def format_items(dct: Dict[str, Any], indent: int) -> None:
        for key, value in dct.items():
            message = " " * indent + f"{key}: "
            if isinstance(value, dict):
                lines.append(message)
                format_items(value, indent + 2)
            else:
                message += str(value)
                lines.append(message)

    format_items(message_dct, 0)

    return "\n".join(lines)


def handle_exception(exc: Exception) -> int:
    """
    Take an exception, print information about it and return the exit code to use
    """
    # TODO: fix this. It's required to avoid a circular import error
    from ggshield.core import ui

    if isinstance(exc, click.exceptions.Abort):
        return ExitCode.SUCCESS

    # Get exit code
    if isinstance(exc, _ExitError):
        exit_code = exc.exit_code
    elif isinstance(exc, (InvalidGitRefError, click.UsageError)):
        exit_code = ExitCode.USAGE_ERROR
    else:
        exit_code = ExitCode.UNEXPECTED_ERROR

    click.echo()
    ui.display_error(str(exc))
    if isinstance(exc, UnicodeEncodeError) and platform.system() == "Windows":
        ui.display_info(
            "\n"
            "ggshield failed to print a message because of an Unicode encoding issue."
            " To workaround that, try setting the PYTHONUTF8 environment variable to 1."
        )

    if not isinstance(exc, (click.ClickException, GitError)):
        click.echo()
        if ui.is_verbose():
            traceback.print_exc()
        else:
            ui.display_info("Re-run the command with --verbose to get a stack trace.")

    return exit_code


def handle_api_error(detail: Detail) -> None:
    # Use %s for status_code because it can be None. Logger is OK with an int being
    # passed for a %s placeholder.
    logger.error("status_code=%s detail=%s", detail.status_code, detail.detail)
    if detail.status_code == 401:
        raise click.UsageError(detail.detail)
    if detail.status_code is None:
        raise UnexpectedError(f"Scanning failed: {detail.detail}")
    if detail.status_code == 403 and detail.detail == "Quota limit reached.":
        raise QuotaLimitReachedError()



import re
import subprocess
import sys
from pathlib import Path
from typing import Pattern, Set
from unittest.mock import Mock, patch

import pytest

from ggshield.cmd.secret.scan.pypi import (
    PYPI_DOWNLOAD_TIMEOUT,
    get_files_from_package,
    save_package_to_tmp,
)
from ggshield.core.errors import UnexpectedError
from ggshield.utils.files import ListFilesMode


class TestPipDownload:
    package_name: str = "what-ever-non-existing"

    def test_pip_download_success(self, tmp_path):
        with patch("subprocess.run") as call:
            save_package_to_tmp(temp_dir=tmp_path, package_name=self.package_name)

            call.assert_called_once_with(
                [
                    "pip",
                    "download",
                    self.package_name,
                    "--dest",
                    str(tmp_path),
                    "--no-deps",
                ],
                check=True,
                stdout=sys.stderr,
                stderr=sys.stderr,
                timeout=PYPI_DOWNLOAD_TIMEOUT,
            )

    def test_pip_download_nonexistent_package(self, tmp_path):
        with patch(
            "subprocess.run", side_effect=subprocess.CalledProcessError(1, cmd=None)
        ):
            with pytest.raises(
                UnexpectedError,
                match=f'Failed to download "{self.package_name}"',
            ):
                save_package_to_tmp(temp_dir=tmp_path, package_name=self.package_name)

    def test_pip_download_timeout(self, tmp_path):
        with patch(
            "subprocess.run",
            side_effect=subprocess.TimeoutExpired(
                cmd=None, timeout=PYPI_DOWNLOAD_TIMEOUT
            ),
        ):
            with pytest.raises(
                UnexpectedError,
                match=(
                    f'Command "pip download {self.package_name} '
                    f'--dest {re.escape(str(tmp_path))} --no-deps" timed out'
                ),
            ):
                save_package_to_tmp(temp_dir=tmp_path, package_name=self.package_name)


class TestListPackageFiles:
    package_name: str = "what-ever-non-existing"
    exclusion_regexes: Set[Pattern[str]] = {re.compile("i am a regex")}

    @pytest.mark.parametrize("extension", ["whl", "tar.gz"])
    @patch("ggshield.cmd.secret.scan.pypi.create_files_from_paths")
    @patch("ggshield.cmd.secret.scan.pypi.safe_unpack")
    def test_unpack_archive_format(
        self,
        safe_unpack_mock: Mock,
        create_files_from_paths_mock: Mock,
        extension: str,
        tmp_path,
    ):
        # TODO: rewrite this test: it tests implementation details instead of the
        # function outcome
        archive_path = tmp_path / f"{self.package_name}.{extension}"
        create_files_from_paths_mock.return_value = ([], [])

        with patch.object(Path, "iterdir", return_value=iter([archive_path])):
            get_files_from_package(
                archive_dir=tmp_path,
                package_name=self.package_name,
                exclusion_regexes=self.exclusion_regexes,
            )

            safe_unpack_mock.assert_called_once_with(
                archive_path,
                extract_dir=tmp_path,
            )

            expected_exclusion_regexes = self.exclusion_regexes
            expected_exclusion_regexes.add(
                re.compile(f"{self.package_name}.{extension}")
            )

            create_files_from_paths_mock.assert_called_once_with(
                paths=[tmp_path],
                exclusion_regexes=expected_exclusion_regexes,
                list_files_mode=ListFilesMode.ALL,
            )



from typing import Any, Optional

import click

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj

from .config_set import set_user_config_field
from .constants import FIELD_NAMES, FIELD_NAMES_DOC, FIELDS


@click.command(
    help=f"""Remove the value of the given configuration key.

If `--all` is passed, iterates over all instances.

{FIELD_NAMES_DOC}
"""
)
@click.argument(
    "field_name",
    nargs=1,
    type=click.Choice(FIELD_NAMES),
    required=True,
    metavar="KEY",
)
@click.option(
    "--instance",
    "instance_url",
    required=False,
    type=str,
    metavar="URL",
    help="Set per instance configuration.",
)
@click.option("--all", "all_", is_flag=True, help="Iterate over all instances.")
@add_common_options()
@click.pass_context
def config_unset_cmd(
    ctx: click.Context,
    field_name: str,
    instance_url: Optional[str],
    all_: bool,
    **kwargs: Any,
) -> int:
    config = ContextObj.get(ctx).config
    field = FIELDS[field_name]

    if field.auth_config:
        if all_:
            setattr(config.auth_config, field_name, None)
            for instance in config.auth_config.instances:
                setattr(instance, field_name, None)

        elif instance_url is None:
            setattr(config.auth_config, field_name, None)

        else:
            instance = config.auth_config.get_instance(instance_url)
            setattr(instance, field_name, None)

        config.auth_config.save()
    else:
        set_user_config_field(field, None)
    return 0



import json
import os
import subprocess
from pathlib import Path
from typing import Dict, Optional, Union

import pytest
from pygitguardian.models import Match

from ggshield.core.filter import censor_match
from tests.functional.conftest import GGSHIELD_PATH


PathLike = Union[Path, str]


def run_ggshield(
    *args: str,
    expected_code: int = 0,
    cwd: Optional[PathLike] = None,
    env: Optional[Dict] = None,
) -> subprocess.CompletedProcess:
    env = env or dict()
    assert GGSHIELD_PATH is not None
    cmd = (GGSHIELD_PATH, *args)
    cwd = None if cwd is None else str(cwd)
    result = subprocess.run(
        cmd,
        check=False,
        text=True,
        capture_output=True,
        cwd=cwd,
        env={**os.environ, **env},
    )
    assert (
        result.returncode == expected_code
    ), f"""Expected returncode {expected_code}, got {result.returncode}

        == stdout ==

        {result.stdout}

        == stderr ==

        {result.stderr}
        """
    return result


def run_ggshield_scan(
    *args: str,
    expected_code: int = 0,
    cwd: Optional[PathLike] = None,
    env: Optional[Dict] = None,
) -> subprocess.CompletedProcess:
    env = env or dict()
    args = ("secret", "scan", *args)
    return run_ggshield(*args, expected_code=expected_code, cwd=cwd, env=env)


def assert_is_valid_json(txt: str) -> None:
    try:
        json.loads(txt)
    except Exception as exc:
        pytest.fail(f"Text is not a valid JSON document:\n---\n{txt}\n---\n{exc}")


def recreate_censored_string(matched_string: str) -> str:
    """Applies ggshield censoring to `matched_string`"""
    match = Match(match=matched_string, match_type="")
    return censor_match(match)


def recreate_censored_content(content: str, matched_string: str) -> str:
    """Applies ggshield censoring to any occurrence of `matched_string` inside
    `content`. Returns the censored string."""
    return content.replace(matched_string, recreate_censored_string(matched_string))



import re
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Any, List, Pattern, Set, Tuple

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.secret.scan.ui_utils import print_file_list
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.errors import UnexpectedError
from ggshield.core.scan import ScanContext, ScanMode, Scannable
from ggshield.core.scan.file import create_files_from_paths
from ggshield.core.scanner_ui import create_scanner_ui
from ggshield.utils.archive import safe_unpack
from ggshield.utils.files import ListFilesMode
from ggshield.verticals.secret import SecretScanCollection, SecretScanner


PYPI_DOWNLOAD_TIMEOUT = 30


def save_package_to_tmp(temp_dir: Path, package_name: str) -> None:
    command: List[str] = [
        "pip",
        "download",
        package_name,
        "--dest",
        str(temp_dir),
        "--no-deps",
    ]

    try:
        ui.display_heading("Downloading package")
        subprocess.run(
            command,
            check=True,
            stdout=sys.stderr,
            stderr=sys.stderr,
            timeout=PYPI_DOWNLOAD_TIMEOUT,
        )

    except subprocess.CalledProcessError:
        raise UnexpectedError(f'Failed to download "{package_name}"')

    except subprocess.TimeoutExpired:
        raise UnexpectedError('Command "{}" timed out'.format(" ".join(command)))


def get_files_from_package(
    archive_dir: Path,
    package_name: str,
    exclusion_regexes: Set[Pattern[str]],
) -> Tuple[List[Scannable], List[Path]]:
    archive: Path = next(archive_dir.iterdir())

    ui.display_heading("Unpacking package")
    try:
        safe_unpack(archive, extract_dir=archive_dir)
    except Exception as exn:
        raise UnexpectedError(f'Failed to unpack package "{package_name}": {exn}.')

    exclusion_regexes.add(re.compile(re.escape(archive.name)))

    return create_files_from_paths(
        paths=[archive_dir],
        exclusion_regexes=exclusion_regexes,
        list_files_mode=ListFilesMode.ALL,
    )


@click.command()
@click.argument("package_name", nargs=1, type=click.STRING, required=True)
@add_secret_scan_common_options()
@click.pass_context
def pypi_cmd(
    ctx: click.Context,
    package_name: str,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan a pypi package.

    Under the hood this command uses the `pip download` command to download the python
    package.

    You can use pip environment variables or configuration files to set `pip download`
    parameters as explained in [pip documentation][1].  For example, you can set pip
    `--index-url` parameter with the `PIP_INDEX_URL` environment variable.

    [1]: https://pip.pypa.io/en/stable/topics/configuration/
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config
    output_handler = create_output_handler(ctx)

    with tempfile.TemporaryDirectory(suffix="ggshield") as temp_dir:
        temp_path = Path(temp_dir)
        save_package_to_tmp(temp_dir=temp_path, package_name=package_name)

        files, binary_paths = get_files_from_package(
            archive_dir=temp_path,
            package_name=package_name,
            exclusion_regexes=ctx_obj.exclusion_regexes,
        )
        print_file_list(files, binary_paths)
        ui.display_heading("Starting scan")

        with create_scanner_ui(len(files)) as scanner_ui:
            scan_context = ScanContext(
                scan_mode=ScanMode.PYPI,
                command_path=ctx.command_path,
            )

            scanner = SecretScanner(
                client=ctx_obj.client,
                cache=ctx_obj.cache,
                secret_config=config.user_config.secret,
                scan_context=scan_context,
            )
            results = scanner.scan(files, scanner_ui=scanner_ui)
        scan = SecretScanCollection(id=package_name, type="path_scan", results=results)

        return output_handler.process_scan(scan)



from dataclasses import dataclass


DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


@dataclass
class ConfigField:
    """Meta-information about a config field, used by `ggshield config` commands to
    manipulate configuration files.
    """

    # The name of the field in the config file.
    name: str

    # True if this field is stored in auth_config.yaml, False if it's in the global
    # user config file.
    auth_config: bool = False

    # True if this field can be set for an individual instance. Only valid for
    # auth_config fields.
    per_instance_ok: bool = False


# All editable config fields
_FIELDS = (
    ConfigField(
        "instance",
    ),
    ConfigField(
        "default_token_lifetime",
        auth_config=True,
        per_instance_ok=True,
    ),
)

FIELDS = {x.name: x for x in _FIELDS}

FIELD_NAMES = sorted(FIELDS.keys())

FIELD_NAMES_DOC = "Supported keys:\n" + "\n".join(f"- `{x}`" for x in FIELD_NAMES)



import tempfile
from pathlib import Path
from typing import Any

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.verticals.secret.docker import docker_save_to_tmp, docker_scan_archive


# bailout if docker command takes longer than 6 minutes
DOCKER_COMMAND_TIMEOUT = 360


@click.command()
@click.option(
    "--docker-timeout",
    type=click.INT,
    default=DOCKER_COMMAND_TIMEOUT,
    help="Timeout for Docker commands.",
    metavar="SECONDS",
    show_default=True,
)
@click.argument("name", nargs=1, type=click.STRING, required=True)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def docker_name_cmd(
    ctx: click.Context, name: str, docker_timeout: int, **kwargs: Any
) -> int:
    """
    Scan a Docker image after exporting its filesystem and manifest with the `docker save` command.

    ggshield tries to pull the image if it's not available locally.
    """

    with tempfile.TemporaryDirectory(suffix="ggshield") as temporary_dir:
        ctx_obj = ContextObj.get(ctx)
        ctx_obj.client = create_client_from_config(ctx_obj.config)
        config = ctx_obj.config
        output_handler = create_output_handler(ctx)

        archive = Path(temporary_dir) / "archive.tar"
        docker_save_to_tmp(name, archive, docker_timeout)

        scan_context = ScanContext(
            scan_mode=ScanMode.DOCKER,
            command_path=ctx.command_path,
        )

        scan = docker_scan_archive(
            archive_path=archive,
            client=ctx_obj.client,
            cache=ctx_obj.cache,
            secret_config=config.user_config.secret,
            scan_context=scan_context,
            exclusion_regexes=ctx_obj.exclusion_regexes,
        )

        return output_handler.process_scan(scan)



import logging
from typing import Any, Dict, Generator, List, Tuple
from urllib.parse import urlparse

import requests

from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.exceptions import (
    VaultForbiddenItemError,
    VaultInvalidUrlError,
    VaultNotFoundItemError,
    VaultPathIsNotADirectoryError,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.models import (
    VaultKvMount,
    VaultSecrets,
)


logger = logging.getLogger(__name__)


class VaultAPIClient:
    """Client to interact with Vault API."""

    def __init__(self, vault_url: str, api_token: str) -> None:
        self.session = requests.Session()
        self.session.headers["X-Vault-Token"] = api_token

        if not vault_url.startswith("http://") and not vault_url.startswith("https://"):
            vault_url = f"https://{vault_url}"
        try:
            self.vault_url = urlparse(vault_url).geturl()
        except ValueError:
            raise VaultInvalidUrlError(
                f"cannot parse the Vault URL '{vault_url}'. Are you sure it is valid?"
            )

    def _make_request(
        self, endpoint: str, method: str = "GET", api_version: str = "v1"
    ) -> Dict[str, Any]:
        """ "
        Make request to the API.
        """
        api_res = self.session.request(
            method, f"{self.vault_url}/{api_version}/{endpoint}"
        )
        api_res.raise_for_status()

        return api_res.json()

    def get_kv_mounts(self) -> Generator[VaultKvMount, None, None]:
        """
        Get all kv mounts for the Vault instance.

        Will try the /sys/mounts endpoint first, and if there is a forbidden error
        fallback to the /sys/internal/ui internal endpoint.
        """

        try:
            api_res = self._make_request("sys/mounts")["data"]
        except requests.HTTPError:
            api_res = self._make_request("sys/internal/ui/mounts")["data"]["secret"]

        for key, value in api_res.items():
            if value["type"] != "kv":
                continue

            yield VaultKvMount(
                name=key.rstrip("/"),  # remove trailing slash
                version=value["options"]["version"],
            )

    def list_kv_items(self, mount: VaultKvMount, path: str) -> List[str]:
        logger.debug(f"Listing kv items for mount {mount.name} at {path}")

        api_endpoint = (
            f"{mount.name}/metadata/{path}"
            if mount.version == "2"
            else f"{mount.name}/{path}"
        )

        try:
            api_res = self._make_request(api_endpoint, method="LIST")
        except requests.HTTPError as exc:
            if exc.response.status_code == 403:
                raise VaultForbiddenItemError(
                    f"cannot access item on mount {mount.name} at path {path}"
                )

            # The API return 404 when trying to list items when the path is a file
            # and not a directory
            if exc.response.status_code == 404:
                raise VaultPathIsNotADirectoryError()

            raise exc

        return [item.rstrip("/") for item in api_res["data"]["keys"]]

    def get_kv_secrets(self, mount: VaultKvMount, path: str) -> List[Tuple[str, str]]:
        """
        Get secrets from the specified mount at the specified path.

        Returns a list of tuples containing secret key and secret value
        """

        logger.debug(f"Getting secrets at {path}")
        api_endpoint = (
            f"{mount.name}/data/{path}"
            if mount.version == "2"
            else f"{mount.name}/{path}"
        )
        try:
            api_res = self._make_request(api_endpoint)
        except requests.HTTPError as exc:
            if exc.response.status_code == 403:
                raise VaultForbiddenItemError(
                    f"cannot access item on mount {mount.name} at path {path}"
                )

            if exc.response.status_code == 404:
                raise VaultNotFoundItemError(
                    f"{path} was not found: it's either not a file, "
                    "was deleted or cannot be accessed with the current token"
                )

            raise exc

        data = api_res["data"]["data"] if mount.version == "2" else api_res["data"]
        return [
            (f"{path}/{secret_name}", secret_value)
            for secret_name, secret_value in data.items()
        ]

    def _get_secrets_or_empty(
        self, mount: VaultKvMount, folder_path: str
    ) -> VaultSecrets:
        """
        Call get_kv_secrets on the given mount and folder path.

        Return the secrets or an empty list if VaultNotFoundItemError was raised.
        """
        try:
            return VaultSecrets(
                secrets=self.get_kv_secrets(mount, folder_path),
                not_fetched_paths=[],
            )
        except VaultNotFoundItemError as exc:
            logger.debug(f"Not found error: {exc}")
            return VaultSecrets(secrets=[], not_fetched_paths=[folder_path])
        except VaultForbiddenItemError as exc:
            logger.debug(f"Forbidden error: {exc}")
            return VaultSecrets(secrets=[], not_fetched_paths=[folder_path])

    def _get_secrets_from_path(
        self, mount: VaultKvMount, folder_path: str, recursive: bool
    ) -> VaultSecrets:
        """
        Get the secrets on the given mount and folder path.
        If recursive is True, iterate recursively on subfolders.

        Return the secrets or an empty list if errors were raised.
        """
        # Get current directory secret
        try:
            subfolders = self.list_kv_items(mount, folder_path)
        except VaultPathIsNotADirectoryError:
            return self._get_secrets_or_empty(mount, folder_path)
        except VaultForbiddenItemError as exc:
            logger.debug(f"Forbidden error: {exc}")
            return VaultSecrets(secrets=[], not_fetched_paths=[folder_path])

        if not recursive:
            return VaultSecrets(secrets=[], not_fetched_paths=[])

        result = VaultSecrets(secrets=[], not_fetched_paths=[])
        for subfolder in subfolders:
            subfolder_result = self._get_secrets_from_path(
                mount, f"{folder_path}/{subfolder}", True
            )
            result.secrets += subfolder_result.secrets
            result.not_fetched_paths += subfolder_result.not_fetched_paths

        return result

    def get_secrets(
        self, mount: VaultKvMount, path: str, recursive: bool
    ) -> VaultSecrets:
        # Check first if it's a directory
        try:
            keys = self.list_kv_items(mount, path)
        except VaultPathIsNotADirectoryError:
            # If it's a file, return directly
            return self._get_secrets_or_empty(mount, path)

        # If it's a folder, get secrets of the folder
        result = VaultSecrets(secrets=[], not_fetched_paths=[])
        for folder in keys:
            folder_path = f"{path}/{folder}".strip("/")
            folder_result = self._get_secrets_from_path(mount, folder_path, recursive)
            result.secrets += folder_result.secrets
            result.not_fetched_paths += folder_result.not_fetched_paths

        return result



import logging
import sys

from ggshield.core.ui.ggshield_ui import (
    NAME_BY_LEVEL,
    DebugInfo,
    GGShieldProgress,
    GGShieldUI,
    Level,
)


class PlainTextGGShieldProgress(GGShieldProgress):
    def advance(self, amount: int) -> None:
        pass


class PlainTextGGShieldUI(GGShieldUI):
    """
    Plain-text implementation of GGShieldUI. Suitable when output is not a TTY.
    """

    def _echo(self, level: Level, message: str) -> None:
        if self.level == Level.DEBUG:
            self._debug_echo(level, message)
        else:
            self._normal_echo(level, message)

    def _debug_echo(self, level: Level, message: str) -> None:
        name = NAME_BY_LEVEL[level][0]
        message = f"[{name}] {message}"
        err(self._prefix() + message)

    def _normal_echo(self, level: Level, message: str) -> None:
        if level <= Level.WARNING:
            name = NAME_BY_LEVEL[level]
            message = f"{name}: {message}"
        err(self._prefix() + message)

    def _echo_heading(self, message: str) -> None:
        prefix = self._prefix()
        err(prefix + message)
        err(prefix + "-" * len(message))

    def log(self, record: logging.LogRecord) -> None:
        msg = f"{record.name}:{record.lineno} {record.getMessage()}"
        level = record.levelno
        if level == logging.DEBUG:
            self.display_debug(msg)
        elif level == logging.INFO:
            self.display_info(msg)
        elif level == logging.WARNING:
            self.display_warning(msg)
        elif level == logging.ERROR:
            self.display_error(msg)
        else:
            self.display_warning(f"Unsupported log level {level}")
            self.display_error(msg)

    def create_progress(self, total: int) -> GGShieldProgress:
        return PlainTextGGShieldProgress()

    def _prefix(self) -> str:
        if self.level < Level.DEBUG:
            return ""
        info = DebugInfo.create()
        return f"{info.timestamp} {info.process_id}:{info.thread_id} "


def err(message: str) -> None:
    """Helper function to print to stderr and flush the output (if we don't flush some
    tests fail)"""
    print(message, file=sys.stderr)
    sys.stderr.flush()



from pathlib import Path
from typing import Any, Dict

import pytest

from ggshield.core.config.utils import (
    find_local_config_path,
    remove_common_dict_items,
    remove_url_trailing_slash,
    replace_dash_in_keys,
    update_dict_from_other,
)
from ggshield.utils.os import cd
from tests.repository import Repository


def test_replace_dash_in_keys():
    """
    GIVEN a dict with some keys using dash
    WHEN replace_dash_in_keys() is called
    THEN the dict keys all use underscore
    AND replace_dash_in_keys() returned a set of all the modified keys
    """
    data = {
        "use_underscore": 12,
        "use-dash": "hello",
        "dash-or-underscore": "dash",
        "dash_or_underscore": "underscore",
        "container": {"sub-dash-key": "values-are-not-affected"},
    }

    dash_keys = replace_dash_in_keys(data)

    assert data == {
        "use_underscore": 12,
        "use_dash": "hello",
        "dash_or_underscore": "underscore",
        "container": {"sub_dash_key": "values-are-not-affected"},
    }
    assert dash_keys == {"use-dash", "sub-dash-key", "dash-or-underscore"}


def test_update_dict_from_other():
    """
    GIVEN two dictionaries
    WHEN update_dict_from_other(dst_conf, src_conf) is called
    THEN dst_conf is updated from src_conf fields
    """

    dst_conf = {
        "subconf": {"an_int": 1},
        "not_overridden_bool": True,
        "overridden_bool": True,
        "a_list": ["i1", "i2"],
        "a_set": {"i1", "i2"},
    }

    src_conf = {
        "overridden_bool": False,
        "subconf": {
            "an_str": "src_subconf_str",
        },
        "a_list": ["i2", "i3"],
        "a_set": {"i2", "i3"},
    }

    update_dict_from_other(dst_conf, src_conf)

    assert dst_conf["not_overridden_bool"] is True
    assert dst_conf["overridden_bool"] is False
    assert dst_conf["subconf"]["an_int"] == 1
    assert dst_conf["subconf"]["an_str"] == "src_subconf_str"
    assert dst_conf["a_list"] == ["i1", "i2", "i2", "i3"]
    assert dst_conf["a_set"] == {"i1", "i2", "i3"}


@pytest.mark.parametrize(
    ["src", "reference", "expected"],
    [
        pytest.param(
            {"modified": 2, "same": 3},
            {"modified": 1, "same": 3, "absent": 4},
            {"modified": 2},
            id="simple",
        ),
        pytest.param(
            {"lst": [3, 4]},
            {"lst": [1, 2]},
            {"lst": [3, 4]},
            id="replace-list",
        ),
        pytest.param(
            {"lst1": [], "lst2": ["a"]},
            {"lst1": [], "lst2": ["a"]},
            {},
            id="identical-lists-are-removed",
        ),
        pytest.param(
            {"lst": []},
            {"lst": ["a"]},
            {"lst": []},
            id="empty-list-not-removed-if-ref-not-empty",
        ),
        pytest.param(
            {"outer": {"inner1": "modified", "inner3": "same"}},
            {"outer": {"inner1": "original", "inner2": "absent", "inner3": "same"}},
            {"outer": {"inner1": "modified"}},
            id="nested-merge",
        ),
        pytest.param(
            {"outer": {"inner1": "same"}},
            {"outer": {"inner1": "same", "inner2": "absent"}},
            {},
            id="nested-dict-is-removed",
        ),
    ],
)
def test_remove_common_dict_items(
    src: Dict[str, Any], reference: Dict[str, Any], expected: Dict[str, Any]
):
    result = remove_common_dict_items(src, reference)
    assert result == expected


def test_remove_url_trailing_slash():
    result = remove_url_trailing_slash("https://dashboard.gitguardian.com/")
    assert result == "https://dashboard.gitguardian.com"


def test_find_config_in_root(tmp_path: Path):
    """
    GIVEN a repo with a config file in the root and a subdirectory
    WHEN trying to find the local config while inside the subdirectory
    THEN the config in the root is returned
    """
    Repository.create(tmp_path)

    config_path = tmp_path / ".gitguardian.yml"
    config_path.touch()

    dir_path = tmp_path / "dir"
    dir_path.mkdir()

    with cd(str(dir_path)):
        assert find_local_config_path() == config_path



import json
from pathlib import Path

import jsonschema
import pytest

from tests.conftest import (
    GG_VALID_TOKEN,
    GG_VALID_TOKEN_IGNORE_SHA,
    KNOWN_SECRET,
    UNKNOWN_SECRET,
)
from tests.functional.utils import (
    recreate_censored_content,
    recreate_censored_string,
    run_ggshield_scan,
)


@pytest.mark.parametrize("path", ("config.py", ".git/config"))
@pytest.mark.parametrize("show_secrets", (True, False))
def test_scan_path(tmp_path: Path, path: str, show_secrets: bool) -> None:
    # GIVEN a secret
    test_file = tmp_path / path
    test_file.parent.mkdir(exist_ok=True, parents=True)
    test_file.write_text(f"SECRET='{GG_VALID_TOKEN}'")

    # WHEN ggshield scans it
    args = ["path", str(test_file)]
    if show_secrets:
        args.append("--show-secrets")
    result = run_ggshield_scan(*args, cwd=tmp_path, expected_code=1)

    # THEN the output contains the context and the expected ignore sha
    assert "SECRET=" in result.stdout
    assert GG_VALID_TOKEN_IGNORE_SHA in result.stdout
    # AND the secrets are shown only with --show-secrets
    if show_secrets:
        assert GG_VALID_TOKEN in result.stdout
    else:
        assert recreate_censored_string(GG_VALID_TOKEN) in result.stdout


def test_scan_path_does_not_fail_on_long_paths(tmp_path: Path) -> None:
    # GIVEN a secret stored in a file whose path is longer than 256 characters
    secret_content = f"SECRET='{GG_VALID_TOKEN}'"

    # Create the file in a subdir because filenames cannot be longer than 255
    # characters. What we care here is the length of the path.
    test_file = tmp_path / ("d" * 255) / ("f" * 255)
    test_file.parent.mkdir()
    test_file.write_text(secret_content)

    # WHEN ggshield scans it
    result = run_ggshield_scan("path", str(test_file), cwd=tmp_path, expected_code=1)

    # THEN it finds the secret in it
    assert recreate_censored_content(secret_content, GG_VALID_TOKEN) in result.stdout


@pytest.mark.parametrize("show_secrets", (True, False))
def test_scan_path_json_output(
    tmp_path: Path, secret_json_schema, show_secrets: bool
) -> None:
    # GIVEN a secret
    test_file = tmp_path / "config.py"
    test_file.write_text(f"SECRET='{GG_VALID_TOKEN}'")

    # WHEN ggshield scans it with --json
    args = ["path", "--json", str(test_file)]
    if show_secrets:
        args.append("--show-secrets")
    result = run_ggshield_scan(*args, cwd=tmp_path, expected_code=1)
    # THEN the output is a valid JSON
    parsed_result = json.loads(result.stdout)
    # AND there is one incident with one occurrence
    assert parsed_result["total_incidents"] == 1
    assert parsed_result["total_occurrences"] == 1
    incident = parsed_result["entities_with_incidents"][0]["incidents"][0]
    # AND it has the expected ignore_sha
    assert incident["ignore_sha"] == GG_VALID_TOKEN_IGNORE_SHA
    # AND the secrets are shown only if --show-secrets has been set
    if show_secrets:
        assert incident["occurrences"][0]["match"] == GG_VALID_TOKEN
    else:
        assert incident["occurrences"][0]["match"] == recreate_censored_string(
            GG_VALID_TOKEN
        )
    # AND the schema is valid
    jsonschema.validate(parsed_result, secret_json_schema)


def test_scan_path_ignore_known_secrets(tmp_path: Path) -> None:
    if not KNOWN_SECRET:
        pytest.fail(
            "You must define $TEST_KNOWN_SECRET to run this test,"
            " see .env.example for details"
        )

    # GIVEN a document containing 2 secrets, one known and one unknown
    test_file = tmp_path / "config.py"
    test_file.write_text(
        f"""
KNOWN_SECRET='{KNOWN_SECRET}'

# Extra lines to ensure the line containing the known secret is not included when
# ggshield shows the lines around the unknown secret.

UNKNOWN_SECRET='{UNKNOWN_SECRET}'
"""
    )

    # WHEN ggshield scans it with --ignore-known-secrets
    result = run_ggshield_scan(
        "path",
        str(test_file),
        "--ignore-known-secrets",
        "--show-secrets",
        cwd=tmp_path,
        expected_code=1,
    )

    # THEN only the unknown secret is reported
    assert f"KNOWN_SECRET='{KNOWN_SECRET}'" not in result.stdout
    assert f"UNKNOWN_SECRET='{UNKNOWN_SECRET}'" in result.stdout



import re
import subprocess
from pathlib import Path
from typing import Dict, List
from unittest.mock import patch

import click
import pytest

from ggshield.core.errors import UnexpectedError
from ggshield.core.scan import File
from ggshield.verticals.secret.docker import (
    DockerImage,
    InvalidDockerArchiveException,
    LayerInfo,
    docker_pull_image,
    docker_save_to_tmp,
)
from tests.unit.conftest import (
    DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH,
    DOCKER_EXAMPLE_LAYER_FILES,
    DOCKER_EXAMPLE_PATH,
)


class ManifestMock:
    def read(self, amount: int = None) -> bytes:
        return b'[{"Config": "8b907fee27ad927c595fcf873c8256796cab27e7a3fb4bf3952308a76ad791c4.json"}]'


class TarMock:
    def __init__(self, members: Dict[str, str], *args, **kwargs):
        self.members = members

    def extractfile(self, member: str):
        if "8b907fee27ad927c595fcf873c8256796cab27e7a3fb4bf3952308a76ad791c4" in member:
            return None
        return self.members.get(member, None)

    def getmember(self, member: str):
        return member if self.members.get(member, None) else None


class TestDockerScan:
    @pytest.mark.parametrize(
        ["op", "want"],
        [
            pytest.param(
                "/bin/sh -c #(nop) COPY dir:xxx in / ",
                True,
            ),
            pytest.param("/bin/sh -c #(nop) ADD dir:xxx in / ", True),
            pytest.param(
                '/bin/sh -c #(nop)  CMD ["/usr/bin/bash"',
                False,
            ),
        ],
    )
    def test_should_scan_layer(self, op: str, want: bool):
        layer_info = LayerInfo(filename="dummy", command=op, diff_id="sha256:1234")
        assert layer_info.should_scan() is want

    @pytest.mark.parametrize(
        ["members", "match"],
        [
            pytest.param({}, "No manifest file found."),
            pytest.param(
                {"manifest.json": ManifestMock()},
                "No config file found.",
            ),
            pytest.param(
                {
                    "manifest.json": ManifestMock(),
                    "8b907fee27ad927c595fcf873c8256796cab27e7a3fb4bf3952308a76ad791c4.json": "layer file",  # noqa: E501
                },  # noqa: E501
                "Config file could not be extracted.",
            ),
        ],
    )
    def test_get_config(self, members, match):
        tarfile = TarMock(members)
        with pytest.raises(InvalidDockerArchiveException, match=match):
            DockerImage(Path("dummy"), tarfile)

    @pytest.mark.parametrize(
        "image_path", [DOCKER_EXAMPLE_PATH, DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH]
    )
    def test_docker_archive(self, image_path: Path):
        with DockerImage.open(image_path) as image:
            # IDs for non-empty layers
            layer_ids: List[str] = []
            # Files for non-empty layers
            layer_files: List[List[File]] = []

            # Fill layer_ids and layer_files
            for info in image.layer_infos:
                if files := list(image.get_layer_scannables(info, set())):
                    layer_ids.append(info.diff_id)
                    layer_files.append(files)

            assert layer_ids == list(DOCKER_EXAMPLE_LAYER_FILES)

            for files, expected_content_dict in zip(
                layer_files, DOCKER_EXAMPLE_LAYER_FILES.values()
            ):
                content_dict = {x.path.as_posix(): x.content for x in files}
                assert content_dict == expected_content_dict


DOCKER_TIMEOUT = 12


class TestDockerPull:
    def test_docker_pull_image_success(self):
        with patch("subprocess.run") as call:
            docker_pull_image("ggshield-non-existant", DOCKER_TIMEOUT)
            call.assert_called_once_with(
                ["docker", "pull", "ggshield-non-existant"],
                check=True,
                timeout=DOCKER_TIMEOUT,
            )

    def test_docker_pull_image_non_exist(self):
        with patch(
            "subprocess.run", side_effect=subprocess.CalledProcessError(1, cmd=[])
        ):
            with pytest.raises(
                click.UsageError,
                match='Image "ggshield-non-existant" not found',
            ):
                docker_pull_image("ggshield-non-existant", DOCKER_TIMEOUT)

    def test_docker_pull_image_timeout(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.TimeoutExpired(cmd=[], timeout=DOCKER_TIMEOUT),
        ):
            with pytest.raises(
                UnexpectedError,
                match='docker pull ggshield-non-existant" timed out',
            ):
                docker_pull_image("ggshield-non-existant", DOCKER_TIMEOUT)


class TestDockerSave:
    TMP_ARCHIVE = Path("/tmp/as/archive.tar")

    def test_docker_save_image_success(self):
        with patch("subprocess.run") as call:
            docker_save_to_tmp(
                "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
            )
            call.assert_called_once_with(
                [
                    "docker",
                    "save",
                    "ggshield-non-existant:latest",
                    "-o",
                    str(self.TMP_ARCHIVE),
                ],
                check=True,
                stderr=-1,
                timeout=DOCKER_TIMEOUT,
            )

    def test_docker_save_image_does_not_exist(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(
                1, cmd=[], stderr=b"reference does not exist"
            ),
        ):
            with pytest.raises(
                click.UsageError,
                match='Image "ggshield-non-existant:latest" not found',
            ):
                docker_save_to_tmp(
                    "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
                )

    def test_docker_save_image_need_pull(self):
        """
        GIVEN a Docker image we do not have locally
        WHEN we try to save it
        THEN we first pull it and then save it

        This test expects the following calls to `docker` commands:

        - docker save <image_name> -o <something>
          -> Fake failure
        - docker pull <image_name>
          -> Fake success
        - docker save <image_name> -o <something>
          -> Fake success
        """
        with patch(
            "subprocess.run",
            side_effect=[
                subprocess.CalledProcessError(
                    1, cmd=[], stderr=b"reference does not exist"
                ),
                None,
                None,
            ],
        ):
            docker_save_to_tmp(
                "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
            )

    def test_docker_save_image_fails(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(
                1, cmd=[], stderr=b"docker failed weirdly"
            ),
        ):
            with pytest.raises(
                UnexpectedError,
                match="Unable to save docker archive:",
            ):
                docker_save_to_tmp(
                    "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
                )

    def test_docker_save_image_timeout(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.TimeoutExpired(cmd=[], timeout=DOCKER_TIMEOUT),
        ):
            expected_msg = f'Command "docker save ggshield-non-existant:latest -o {str(self.TMP_ARCHIVE)}" timed out'  # noqa: E501
            with pytest.raises(
                UnexpectedError,
                match=re.escape(expected_msg),
            ):
                docker_save_to_tmp(
                    "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
                )



import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable, List, Optional, Pattern, Set, Tuple

from ggshield.utils.files import is_path_excluded
from ggshield.utils.git_shell import Filemode, git
from ggshield.utils.itertools import batched

from .scannable import Scannable


# Used instead of the SHA in commit URLs for staged changes
STAGED_PREFIX = "staged"

# Used instead of the SHA in commit URLs for commit files created from a patch
PATCH_PREFIX = "patch"

# Command line arguments passed to `git show` and `git diff` to get parsable headers
HEADER_COMMON_ARGS = [
    "--raw",  # shows a header with the files touched by the commit
    "-z",  # separate file names in the raw header with \0
]

# Command line arguments passed to `git show` and `git diff` to get parsable patches
PATCH_COMMON_ARGS = [
    *HEADER_COMMON_ARGS,
    "--patch",  # force output of the diff (--raw disables it)
]

# This block is inserted before the patch by commands generating patches from the staging
# area, where there is no commit info yet
DIFF_EMPTY_COMMIT_INFO_BLOCK = """Author:   <>\nDate:  \n:"""

_RX_HEADER_FILE_LINE_SEPARATOR = re.compile("[\n\0]:", re.MULTILINE)

# Match the path in a "---a/file_path" or a "+++ b/file_path".
# Note that for some reason, git sometimes append an \t at the end (happens with the
# "I'm unusual!" file in the test suite). We ignore it.
OLD_NAME_RX = re.compile(r"^--- a/(.*?)\t?$", flags=re.MULTILINE)
NEW_NAME_RX = re.compile(r"^\+\+\+ b/(.*?)\t?$", flags=re.MULTILINE)

MULTI_PARENT_HUNK_HEADER_RX = re.compile(
    r"^(?P<at>@@+) (?P<from>-\d+(?:,\d+)?) .* (?P<to>\+\d+(?:,\d+)?) @@+(?P<trailing_content>.+)?"
)

MAX_FILES_PER_GIT_COMMAND = 100


class PatchParseError(Exception):
    """
    Raised by parse_patch() if it fails to parse its patch.
    """

    pass


class CommitScannable(Scannable):
    """Represents a file inside a commit. The URL of a CommitScannable looks like
    this:

        commit://<sha>/<path>

    For staged commits and commits from patches, `<sha>` is replaced with STAGED_PREFIX
    and PATCH_PREFIX respectively.
    """

    def __init__(
        self,
        sha: Optional[str],
        path: Path,
        content: str,
        filemode: Filemode = Filemode.MODIFY,
    ) -> None:
        super().__init__(filemode)
        self._sha = sha
        self._path = path
        self._content = content
        self._utf8_encoded_size = None

    def _read_content(self) -> None:
        if self._content is not None and self._utf8_encoded_size is None:
            self._utf8_encoded_size = len(self._content.encode(errors="replace"))

    @property
    def url(self) -> str:
        return CommitScannable.create_url(self._sha, self.path)

    @property
    def filename(self) -> str:
        return self.url

    @property
    def path(self) -> Path:
        return self._path

    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        self._read_content()
        assert self._utf8_encoded_size is not None
        return self._utf8_encoded_size > max_utf8_encoded_size

    @staticmethod
    def create_url(sha: Optional[str], path: Path) -> str:
        """Creates a Commit URL based on the sha and Path. It is exposed as a static
        method so that code working with not-yet-parsed commits can return an URL for
        them."""
        prefix = sha if sha else STAGED_PREFIX
        return f"commit://{prefix}/{path.as_posix()}"


@dataclass
class PatchFileInfo:
    """
    Stores information about a file modified by a patch
    """

    # old_path is None unless filemode is RENAME or COPY
    old_path: Optional[Path]
    path: Path
    mode: Filemode

    @staticmethod
    def from_string(line: str) -> "PatchFileInfo":
        """
        Parse a file line in the raw patch header, returns a PatchFileInfo

        See https://github.com/git/git/blob/master/Documentation/diff-format.txt for
        details on the format.
        """

        prefix, path, *rest = line.rstrip("\0").split("\0")

        if rest:
            # If the line has a new path, it's a rename
            old_path = Path(path)
            new_path = Path(rest[0])
        else:
            old_path = None
            new_path = Path(path)

        # for a non-merge commit, prefix is
        # :old_perm new_perm old_sha new_sha status_and_score
        #
        # for a 2 parent commit, prefix is
        # ::old_perm1 old_perm2 new_perm old_sha1 old_sha2 new_sha status_and_score
        #
        # We can ignore most of it, because we only care about the status.
        #
        # status_and_score is one or more status letters, followed by an optional
        # numerical score. We can ignore the score, but we need to check the status
        # letters.
        status = prefix.rsplit(" ", 1)[-1].rstrip("0123456789")

        # There is one status letter per commit parent. In the case of a non-merge
        # commit the situation is simple: there is only one letter.
        # In the case of a merge commit we must look at all letters: if one parent is
        # marked as D(eleted) and the other as M(odified) then we use MODIFY as filemode
        # because the end result contains modifications. To ensure this, the order of
        # the `if` below matters.

        if "M" in status:  # modify
            mode = Filemode.MODIFY
        elif "C" in status:  # copy
            mode = Filemode.NEW
        elif "A" in status:  # add
            mode = Filemode.NEW
        elif "T" in status:  # type change
            mode = Filemode.NEW
        elif "R" in status:  # rename
            mode = Filemode.RENAME
        elif "D" in status:  # delete
            mode = Filemode.DELETE
        else:
            raise ValueError(f"Can't parse header line {line}: unknown status {status}")

        return PatchFileInfo(old_path, new_path, mode)


@dataclass
class PatchHeader:
    """
    Semi-parsed information about a patch. `info` is not parsed because some functions
    do not need the details in it.
    """

    # Meta information: commit author, date, message
    info: str

    # Files added/modified/removed by the commit
    files: List[PatchFileInfo] = field(default_factory=list)

    @staticmethod
    def from_string(header: str) -> "PatchHeader":
        # First item returned by split() contains commit info and message, skip it
        info, *lines = _RX_HEADER_FILE_LINE_SEPARATOR.split(header)
        return PatchHeader(
            info,
            [PatchFileInfo.from_string(x) for x in lines],
        )


def parse_patch(
    sha: Optional[str], patch: str, exclusion_regexes: Optional[Set[Pattern[str]]]
) -> Iterable[Scannable]:
    """
    Parse a patch generated with `git show` or `git diff` using PATCH_COMMON_ARGS.
    Returns a list of Scannable.

    A patch looks like this:

    ```
    commit $SHA
    Author: $NAME <$EMAIL>
    Date:   $DATE

        $SUBJECT

        $BODY1
        $BODY2
        ...

    $AFFECTED_FILE_LINE\0$DIFF1
    $DIFF2
    ...
    ```

    For a non-merge commit, $DIFFn looks like this:

    ```
    diff --git $A_NAME $B_NAME
    $META_INFO
    $META_INFO
    ...
    --- $A_NAME
    +++ $A_NAME
    @@ $FROM $TO @@
    $ACTUAL_CHANGES
    @@ $FROM $TO @@
    $MORE_CHANGES
    ```

    $A_NAME and $B_NAME may be /dev/null in case of creation or removal. When they are
    not, they start with "a/" and "b/" respectively.

    For a 2-parent merge commit with resolved conflicts, $DIFFn looks like this:

    ```
    diff --cc $NAME
    $META_INFO
    $META_INFO
    ...
    --- $A_NAME
    +++ $B_NAME
    @@@ $FROM1 $FROM2 $TO @@@
    $ACTUAL_CHANGES
    ```

    Note that:
    - The diff line only contains one name, without any "a/" or "b/" prefixes.
    - The hunk starts with 3 "@" instead of 2. For a commit with N parents, there are
      actually N+1 "@" characters.
    """
    if exclusion_regexes is None:
        exclusion_regexes = set()

    tokens = patch.split("\0diff ", 1)
    if len(tokens) == 1:
        # No diff, we are done
        return
    header_str, rest = tokens

    try:
        header = PatchHeader.from_string(header_str)

        diffs = re.split(r"^diff ", rest, flags=re.MULTILINE)
        for diff in diffs:
            # Split diff into header and content
            try:
                # + 1 because we match the "\n" in "\n@@"
                content_start = diff.index("\n@@") + 1
            except ValueError:
                # No content
                continue
            diff_header = diff[:content_start]
            content = diff[content_start:]

            # Find diff path in diff header
            match = NEW_NAME_RX.search(diff_header)
            if not match:
                # Must have been deleted. find the old path in this case
                match = OLD_NAME_RX.search(diff_header)
                if not match:
                    raise PatchParseError(
                        f"Could not find old path in {repr(diff_header)}"
                    )
            path = Path(match.group(1))
            if is_path_excluded(path, exclusion_regexes):
                continue

            file_info = next(x for x in header.files if x.path == path)

            if content.startswith("@@@"):
                content = convert_multi_parent_diff(content)

            yield CommitScannable(sha, file_info.path, content, filemode=file_info.mode)
    except Exception as exc:
        if sha:
            msg = f"Could not parse patch (sha: {sha}): {exc}"
        else:
            msg = f"Could not parse patch: {exc}"
        raise PatchParseError(msg)


def convert_multi_parent_diff(content: str) -> str:
    """
    ggshield output handlers currently do not work with multi-parent diffs, so convert
    them into single-parent diffs
    """
    lines = content.splitlines()

    # Process header
    hunk_header, parent_count = process_multi_parent_hunk_header(lines.pop(0))
    out_lines = [hunk_header]

    # Process content
    for line in lines:
        columns = line[:parent_count]
        text = line[parent_count:]
        if columns.startswith("-"):
            # Removed from first parent, keep it
            text = f"-{text}"
        elif columns.startswith("+"):
            # Added by first parent, keep it
            text = f"+{text}"
        elif "+" in columns:
            # Added by another parent, keep it but consider it unchanged
            text = f" {text}"
        elif "-" in columns:
            # Removed from another parent, skip it
            continue
        else:
            # Unchanged
            text = f" {text}"
        out_lines.append(text)

    return "\n".join(out_lines)


def process_multi_parent_hunk_header(header: str) -> Tuple[str, int]:
    match = MULTI_PARENT_HUNK_HEADER_RX.match(header)
    if not match:
        raise PatchParseError(  # pragma: no cover
            f"Failed to parse multi-parent hunk header '{header}'"
        )

    from_ = match.group("from")
    to = match.group("to")
    new_hunk_header = f"@@ {from_} {to} @@"
    if match.group("trailing_content"):
        new_hunk_header += f"{match.group('trailing_content')}"

    # Parent count is the number of '@' at the beginning of the header, minus 1
    parent_count = len(match.group("at")) - 1

    return new_hunk_header, parent_count


def get_file_sha_in_ref(
    ref: str,
    files: List[str],
    cwd: Optional[Path] = None,
) -> Iterable[Tuple[str, str]]:
    """
    Helper function to get the shas of files in the git reference.
    """
    for files in batched(files, MAX_FILES_PER_GIT_COMMAND):
        output = git(["ls-tree", "-z", ref] + files, cwd=cwd)
        for line in output.split("\0")[:-1]:
            _, _, sha, path = line.split(maxsplit=3)
            yield (path, sha)


def get_file_sha_stage(
    files: List[str], cwd: Optional[Path] = None
) -> Iterable[Tuple[str, str]]:
    """
    Helper function to get the shas currently staged of files.
    """
    for files in batched(files, MAX_FILES_PER_GIT_COMMAND):
        output = git(["ls-files", "--stage", "-z"] + files, cwd=cwd)
        for line in output.split("\0")[:-1]:
            _, sha, _, path = line.split(maxsplit=3)
            yield (path, sha)


def get_diff_files(cwd: Optional[Path] = None) -> List[str]:
    """
    Helper function to get the files modified and staged.
    """
    output = git(["diff", "--name-only", "--staged", "-z"], cwd=cwd)
    return output.split("\0")[:-1]  # remove the trailing \0



from typing import Any, Sequence

from typing_extensions import Self

from ggshield.core.scan import Scannable
from ggshield.core.scanner_ui.scanner_ui import ScannerUI
from ggshield.core.ui.ggshield_ui import GGShieldUI


class RichMessageOnlyScannerUI(ScannerUI):
    """
    Basic UI, only supports showing messages when `on_*()` methods are called
    """

    def __init__(self, ui: GGShieldUI):
        self.ui = ui

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any) -> None:
        pass

    def on_scanned(self, scannables: Sequence[Scannable]) -> None:
        for scannable in scannables:
            self.ui.display_verbose(f"Scanned {scannable.path}")

    def on_skipped(self, scannable: Scannable, reason: str) -> None:
        if reason:
            message = f"Skipped {scannable.path}: {reason}"
            self.ui.display_info(message)


class RichProgressScannerUI(RichMessageOnlyScannerUI):
    """
    Show a progress bar in addition to messages when `on_*()` methods are called
    """

    def __init__(self, ui: GGShieldUI, total: int):
        super().__init__(ui)
        self.progress = ui.create_progress(total)

    def __enter__(self) -> Self:
        self.progress.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        self.progress.__exit__(*args)

    def on_scanned(self, scannables: Sequence[Scannable]) -> None:
        super().on_scanned(scannables)
        self.progress.advance(len(scannables))

    def on_skipped(self, scannable: Scannable, reason: str) -> None:
        super().on_skipped(scannable, reason)
        self.progress.advance(1)



from pygitguardian.models import Match

from ggshield.core.lines import Line, get_lines_from_content
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret.extended_match import ExtendedMatch
from tests.unit.conftest import _SINGLE_MOVE_PATCH


def test_censor_match_for_a_patch():
    """
    GIVEN a match against a patch
    AND its lines
    WHEN ExtendedMatch.from_match()  and ext.censor() are called
    THEN it contains the expected values
    """

    # When scanning a patch we send the content after the first "@@" of the patch
    content = _SINGLE_MOVE_PATCH[_SINGLE_MOVE_PATCH.index("@@") :]
    lines = get_lines_from_content(content, Filemode.MODIFY)
    start = 40
    end = 108
    match = Match(
        match=content[start : end + 1],
        match_type="token",
        index_start=start,
        index_end=end,
    )

    ex_match = ExtendedMatch.from_match(match, lines, is_patch=True)
    len_match = len(ex_match.match)
    ex_match.censor()

    assert len(ex_match.match) == len_match
    assert (
        ex_match.match
        == "SG._Yytrtvlj******************************************-**rRJLGFLBLf0M"
    )
    assert ex_match.lines_before_secret == [
        Line(
            content="@@ -150 +150,2 @@",
            is_patch=True,
            pre_index=None,
            post_index=None,
        ),
        Line(
            content="+something",
            is_patch=True,
            pre_index=None,
            post_index=150,
        ),
    ]
    assert ex_match.lines_with_secret == [
        Line(
            content=' sg_key = "SG._Yytrtvlj******************************************-**rRJLGFLBLf0M";',
            is_patch=True,
            pre_index=150,
            post_index=151,
        )
    ]


PLAIN_CONTENT_MUTIPLE_ON_ONE_LINE = """/*
01234567890
*/
token1=ABCD token2=EFGH

# Some more content
# Even more content
"""


def test_censor_match_for_plain_content_multiple_on_one_line():
    """
    GIVEN some content
    AND a match in it
    WHEN ExtendedMatch.from_match() is called
    THEN it contains the expected values
    """
    lines = get_lines_from_content(PLAIN_CONTENT_MUTIPLE_ON_ONE_LINE, Filemode.FILE)
    start1 = PLAIN_CONTENT_MUTIPLE_ON_ONE_LINE.find("A")
    # No + 1 here because in a Match object, index_end points to the last character of
    # the match, not the character after it.
    end1 = PLAIN_CONTENT_MUTIPLE_ON_ONE_LINE.find("D")
    match1 = Match(match="ABCD", match_type="token", index_start=start1, index_end=end1)
    start2 = PLAIN_CONTENT_MUTIPLE_ON_ONE_LINE.find("E")
    end2 = PLAIN_CONTENT_MUTIPLE_ON_ONE_LINE.find("H")
    match2 = Match(match="EFGH", match_type="token", index_start=start2, index_end=end2)

    ex_match1 = ExtendedMatch.from_match(match1, lines, is_patch=False)
    ex_match2 = ExtendedMatch.from_match(match2, lines, is_patch=False)
    ex_match1.censor()
    ex_match2.censor()

    assert (
        ex_match1.lines_before_secret
        == ex_match2.lines_before_secret
        == [
            Line(
                content="01234567890",
                is_patch=False,
                pre_index=2,
            ),
            Line(content="*/", is_patch=False, pre_index=3),
        ]
    )
    assert (
        ex_match1.lines_after_secret
        == ex_match2.lines_after_secret
        == [
            Line(content="", is_patch=False, pre_index=5),
            Line(content="# Some more content", is_patch=False, pre_index=6),
        ]
    )
    assert (
        ex_match1.lines_with_secret
        == ex_match2.lines_with_secret
        == [
            Line(
                content="token1=A**D token2=E**H",
                is_patch=False,
                pre_index=4,
            ),
        ]
    )



from typing import Optional

import pytest

from ggshield.core import ui
from ggshield.core.ui import Level


def test_level():
    assert ui.get_level() == Level.INFO

    ui.set_level(Level.VERBOSE)
    assert ui.get_level() == Level.VERBOSE


@pytest.mark.parametrize(
    ("level", "expected"),
    (
        (None, False),
        (Level.VERBOSE, True),
        (Level.DEBUG, True),
        (Level.ERROR, False),
    ),
)
def test_is_verbose(level: Optional[Level], expected: bool):
    if level:
        ui.set_level(level)
    assert ui.is_verbose() == expected



import os
from typing import List, Tuple

from ggshield.core.errors import UnexpectedError
from ggshield.core.git_hooks.ci.get_scan_ci_parameters import (
    CI_TARGET_BRANCH_ASSOC,
    get_remote_prefix,
)
from ggshield.utils.git_shell import EMPTY_SHA, get_list_commit_SHA

from ... import ui
from .supported_ci import SupportedCI


CI_PREVIOUS_COMMIT_VAR = {
    SupportedCI.JENKINS: "GIT_PREVIOUS_COMMIT",
    SupportedCI.CIRCLECI: "CIRCLE_RANGE",
    SupportedCI.TRAVIS: "TRAVIS_COMMIT_RANGE",
    SupportedCI.GITLAB: "CI_COMMIT_BEFORE_SHA",
    SupportedCI.GITHUB: "GITHUB_PUSH_BASE_SHA",
    SupportedCI.DRONE: "DRONE_COMMIT_BEFORE",
}

CI_COMMIT_VAR = {
    SupportedCI.JENKINS: "GIT_COMMIT",
    SupportedCI.CIRCLECI: "CIRCLE_SHA1",
    SupportedCI.TRAVIS: "TRAVIS_COMMIT",
    SupportedCI.GITLAB: "CI_COMMIT_SHA",
    SupportedCI.GITHUB: "GITHUB_SHA",
    SupportedCI.AZURE: "BUILD_SOURCEVERSION",
    SupportedCI.BITBUCKET: "BITBUCKET_COMMIT",
}


def collect_commit_range_from_ci_env() -> Tuple[List[str], SupportedCI]:
    ci_mode = SupportedCI.from_ci_env()

    base_commit_var = CI_COMMIT_VAR.get(ci_mode)
    base_commit = os.getenv(base_commit_var, "HEAD") if base_commit_var else "HEAD"

    ui.display_verbose(f"\tIdentified base commit as {base_commit}")

    target_branch_var = CI_TARGET_BRANCH_ASSOC.get(ci_mode)
    target_branch = None
    if target_branch_var:
        target_branch = os.getenv(target_branch_var)
        if target_branch and target_branch != EMPTY_SHA:
            ui.display_verbose(f"\tIdentified target branch as {target_branch}")
            commit_list = get_list_commit_SHA(
                f"{get_remote_prefix()}{target_branch}..{base_commit}"
            )
            if commit_list:
                return commit_list, ci_mode

    previous_commit = None
    previous_commit_var = CI_PREVIOUS_COMMIT_VAR.get(ci_mode)
    if previous_commit_var:
        previous_commit = os.getenv(previous_commit_var)
        if (
            previous_commit is None or previous_commit == EMPTY_SHA
        ) and ci_mode == SupportedCI.GITHUB:
            previous_commit = os.getenv("GITHUB_DEFAULT_BRANCH")
        if (
            previous_commit is not None
            and previous_commit != EMPTY_SHA
            and not previous_commit.startswith("...")
        ):
            ui.display_verbose(
                f"\tIdentified previous commit or commit range as {previous_commit}"
            )
            if ci_mode in [SupportedCI.CIRCLECI, SupportedCI.TRAVIS]:
                # for these ci envs, previous_commit is a range of commits
                commit_range = previous_commit
            elif ci_mode == SupportedCI.GITLAB:
                commit_range = f"{previous_commit}~1..{base_commit}"
            else:
                commit_range = f"{previous_commit}..{base_commit}"
            commit_list = get_list_commit_SHA(commit_range)
            if commit_list:
                return commit_list, ci_mode

    commit_list = get_list_commit_SHA(f"{base_commit}~1...")
    if commit_list:
        return commit_list, ci_mode

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"  CI_TYPE: {ci_mode.value}\n"
        f"  TARGET_BRANCH: {target_branch}\n"
        f"  PREVIOUS_COMMIT: {previous_commit}\n"
        f"  BASE_COMMIT: {base_commit}"
    )



from typing import Any, Dict, List, Union

import pytest
from click.testing import CliRunner
from pytest_voluptuous import S

from ggshield.__main__ import cli
from ggshield.core.errors import ExitCode
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok, my_vcr
from tests.unit.request_mock import RequestMock, create_json_response


def test_honeytokens_create_no_arg(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN -
    WHEN running the honeytoken command with no argument
    THEN the return code is 2
    """

    result = cli_fs_runner.invoke(
        cli,
        [
            "honeytoken",
            "create",
        ],
    )
    assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)


@my_vcr.use_cassette("test_honeytoken_create_ok_no_name")
def test_honeytoken_create_ok_no_name(cli_fs_runner: CliRunner) -> None:
    """
    GIVEN -
    WHEN running the honeytoken command with no name
    THEN the return code is 0 and a name is generated
    """

    result = cli_fs_runner.invoke(
        cli,
        ["honeytoken", "create", "--description", "description", "--type", "AWS"],
    )
    assert_invoke_ok(result)


def test_honeytoken_create_ok(cli_fs_runner: CliRunner, monkeypatch) -> None:
    """
    GIVEN -
    WHEN running the honeytoken command with all needed arguments
    THEN the return code is 0
    """
    type_ = "AWS"
    name = "test_honey_token"
    description = "description"
    mock = RequestMock()
    monkeypatch.setattr("ggshield.core.client.Session.request", mock)

    def payload_checker(body: Union[List[str], Dict[str, Any]]) -> None:
        assert (
            S(
                {
                    "type": type_,
                    "description": description,
                    "name": name,
                }
            )
            == body
        )

    mock.add_POST(
        "/honeytokens",
        create_json_response(
            {
                "id": "858a3a76-f6f2-4c2e-b0bc-123456789012",
                "type": type_,
                "status": "active",
                "created_at": "2023-05-15T14:09:09.210845Z",
                "gitguardian_url": "https://dashboard.gitguardian.com/workspace/1/honeytokens/858a3a76-f6f2-4c2e-b0bc-123456789012",  # noqa: E501
                "revoked_at": None,
                "triggered_at": None,
                "open_events_count": 0,
                "creator_id": 265476,
                "creator_api_token_id": "8219e02d-f44c-4802-8418-210987654321",
                "revoker_id": None,
                "revoker_api_token_id": None,
                "token": {
                    "access_token_id": "ABCQFRJEIGOERJ5TRMP",
                    "secret_key": "9ImlMcdJrfjkriegj3454566C0YgLEgregerZEaa",  # ggignore
                },
                "tags": [],
                "name": name,
                "description": description,
            },
            201,
        ),
        json_checker=payload_checker,
    )

    result = cli_fs_runner.invoke(
        cli,
        [
            "honeytoken",
            "create",
            "--description",
            description,
            "--type",
            type_,
            "--name",
            name,
        ],
    )
    assert_invoke_ok(result)


@pytest.mark.parametrize(
    "subcommand, endpoint",
    [
        ("create", "/honeytokens"),
        ("create-with-context", "/honeytokens/with-context"),
    ],
)
@pytest.mark.parametrize(
    "response_message, expected_stdout",
    [
        (
            "Token is missing the following scope: honeytokens:read",
            "ggshield does not have permissions to create honeytokens",
        ),
        (
            "The account has an IP allowlist enabled, and your IP address "
            "is not permitted to access this resource.",
            "The account has an IP allowlist enabled, and your IP address "
            "is not permitted to access this resource.",
        ),
    ],
)
def test_honeytoken_create_error_403(
    cli_fs_runner: CliRunner,
    monkeypatch,
    subcommand: str,
    endpoint: str,
    response_message: str,
    expected_stdout: str,
) -> None:
    """
    GIVEN a command that will cause a 403 status code
    WHEN running the honeytoken command
    THEN the return code is UNEXPECTED_ERROR and the error message matches the expected message
    """

    mock = RequestMock()
    monkeypatch.setattr("ggshield.core.client.Session.request", mock)

    def payload_checker(body: Union[List[str], Dict[str, Any]]) -> None:
        assert body["type"] == "AWS"

    mock.add_POST(
        endpoint,
        create_json_response(
            {"detail": response_message},
            403,
        ),
        json_checker=payload_checker,
    )

    result = cli_fs_runner.invoke(
        cli,
        [
            "honeytoken",
            subcommand,
            "--type",
            "AWS",
        ],
    )
    assert_invoke_exited_with(result, ExitCode.UNEXPECTED_ERROR)
    mock.assert_all_requests_happened()
    assert expected_stdout in result.stdout



from pathlib import Path
from typing import Optional
from zipfile import ZipFile

import pytest

from ggshield.utils.archive import (
    UnsafeArchive,
    _zip_get_symlink_target,
    check_archive_content,
    safe_unpack,
)
from tests.unit.conftest import DATA_PATH


ARCHIVES_PATH = DATA_PATH / "archives"
BAD_ZIP_PATH = ARCHIVES_PATH / "bad.zip"
BAD_TAR_PATH = ARCHIVES_PATH / "bad.tar"
BAD_JAR_PATH = ARCHIVES_PATH / "bad.jar"
GOOD_ZIP_PATH = ARCHIVES_PATH / "good.zip"
GOOD_WHL_PATH = ARCHIVES_PATH / "good.whl"
GOOD_TAR_PATH = ARCHIVES_PATH / "good.tar"
GOOD_JAR_PATH = ARCHIVES_PATH / "good.jar"

"""
Both bad.zip, bad.tar, and bad.jar have the same content:

./
./fine
./subdir/
./subdir/bad-relative-symlink -> ../../bad-relative
./subdir/bad-absolute-symlink -> /tmp/bad-absolute
./subdir/fine-symlink -> ../fine
../bad-relative
/tmp/bad-absolute
"""


@pytest.mark.parametrize("archive", [BAD_ZIP_PATH, BAD_TAR_PATH, BAD_JAR_PATH])
def test_check_archive_content_raises_exception(archive: Path):
    """
    GIVEN a bad archive
    WHEN check_archive_content() is called
    THEN it raises an exception
    AND the exception contains the description of what's bad in the archive
    """
    with pytest.raises(UnsafeArchive) as exc_info:
        check_archive_content(archive)

    bad_paths = set(exc_info.value.bad_paths)
    assert bad_paths == {Path("../bad-relative"), Path("/tmp/bad-absolute")}

    bad_links = set(exc_info.value.bad_links)
    assert bad_links == {
        (Path("./subdir/bad-relative-symlink"), Path("../../bad-relative")),
        (Path("./subdir/bad-absolute-symlink"), Path("/tmp/bad-absolute")),
    }

    # Basic message check, ensures __str__() is exercised
    message = str(exc_info.value)
    assert "Paths outside the archive root:" in message
    assert "bad-relative" in message
    assert "Links pointing outside the archive root:" in message
    assert "bad-absolute-symlink" in message


@pytest.mark.parametrize(
    "archive", [GOOD_ZIP_PATH, GOOD_WHL_PATH, GOOD_TAR_PATH, GOOD_JAR_PATH]
)
def test_check_safe_unpack(tmp_path: Path, archive: Path):
    """
    GIVEN a good archive
    WHEN safe_unpack() is called
    THEN it unpacks the archive
    """
    safe_unpack(archive, tmp_path)
    unpacked_files = set(p for p in tmp_path.rglob("*") if not p.is_dir())
    assert unpacked_files == {tmp_path / "fine", tmp_path / "subdir/fine-symlink"}


@pytest.mark.parametrize(
    ("name", "target"),
    [
        ("./subdir/fine-symlink", Path("../fine")),
        ("./fine", None),
    ],
)
def test_zip_get_symlink_target(name: str, target: Optional[Path]):
    """
    GIVEN a path inside bad.zip
    WHEN asked for its symlink target
    THEN it returns the target if it's a symlink, or None if it's a regular file
    """
    with ZipFile(BAD_ZIP_PATH) as zip:
        info = zip.getinfo(name)
        assert _zip_get_symlink_target(zip, info) == target



from pathlib import Path

from click.testing import CliRunner

from ggshield.__main__ import cli
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok, my_vcr


@my_vcr.use_cassette
def test_hmsl_query_prefix(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN a common prefix
    WHEN running the check command on it
    THEN we should find a match
    """
    payload_path = tmp_path / "payload.txt"
    payload_path.write_text("743d9")

    result = cli_fs_runner.invoke(cli, ["hmsl", "query", str(payload_path)])

    assert "password" not in result.output  # encrypted
    assert "payload" in result.output
    assert_invoke_ok(result)


@my_vcr.use_cassette
def test_hmsl_query_hash(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN a common hash
    WHEN running the check command on it
    THEN we should find a match and no location URL is returned
    """
    payload_path = tmp_path / "payload.txt"
    payload_path.write_text(
        "743d9fde380b7064cc6a8d3071184fc47905cf7440e5615cd46c7b6cbfb46d47"
    )
    result = cli_fs_runner.invoke(cli, ["hmsl", "query", str(payload_path)])
    assert_invoke_ok(result)
    assert "github.com" not in result.output  # already decrypted


def test_bad_payload(cli_fs_runner: CliRunner, tmp_path: Path) -> None:
    """
    GIVEN an invalid payload
    WHEN running the check command on it
    THEN we should get an error
    """
    payload_path = tmp_path / "payload.txt"
    payload_path.write_text("foo")
    result = cli_fs_runner.invoke(cli, ["hmsl", "query", str(payload_path)])
    assert_invoke_exited_with(result, 128)



from typing import Any, Sequence

from typing_extensions import Self

from ggshield.core import ui
from ggshield.core.scan import Scannable
from ggshield.core.scanner_ui.scanner_ui import ScannerUI


class PlainTextScannerUI(ScannerUI):
    """
    Plain-text implementation of ScannerUI. Does not show progress.
    """

    def on_scanned(self, scannables: Sequence[Scannable]) -> None:
        pass

    def on_skipped(self, scannable: Scannable, reason: str) -> None:
        if reason:
            ui.display_info(f"Skipped {scannable.url}: {reason}")

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any) -> None:
        pass



from pathlib import Path
from typing import Any

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.click import RealPath
from ggshield.verticals.secret.docker import docker_scan_archive


@click.command(hidden=True)
@click.argument(
    "archive", nargs=1, type=RealPath(exists=True, resolve_path=True), required=True
)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def docker_archive_cmd(
    ctx: click.Context,
    archive: Path,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan a docker archive <ARCHIVE> without attempting to save or pull the image.

    Hidden command `ggshield secret scan docker-archive`
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config
    output_handler = create_output_handler(ctx)

    scan_context = ScanContext(
        scan_mode=ScanMode.DOCKER,
        command_path=ctx.command_path,
    )

    scan = docker_scan_archive(
        archive_path=archive,
        client=ctx_obj.client,
        cache=ctx_obj.cache,
        secret_config=config.user_config.secret,
        scan_context=scan_context,
        exclusion_regexes=ctx_obj.exclusion_regexes,
    )

    return output_handler.process_scan(scan)



import os
import tempfile
from unittest.mock import Mock, patch

from pygitguardian.models import Match, PolicyBreak

from ggshield.__main__ import cli
from ggshield.cmd.secret.ignore import ignore_last_found
from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.errors import ExitCode
from ggshield.core.filter import get_ignore_sha
from ggshield.core.scan import Commit, ScanContext, ScanMode
from ggshield.core.types import IgnoredMatch
from ggshield.verticals.secret import SecretScanner
from tests.unit.conftest import (
    _MULTIPLE_SECRETS_PATCH,
    assert_invoke_exited_with,
    assert_invoke_ok,
    my_vcr,
)


DOT_GITGUARDIAN_YAML = os.path.join(tempfile.gettempdir(), ".gitguardian.yml")

FOUND_SECRETS = [
    IgnoredMatch(
        name="MySQL Assignment - test.txt",
        match="41b8889e5e794b21cb1349d8eef1815960bf5257330fd40243a4895f26c2b5c8",
    )
]


def compare_matches_ignore(match):
    return (match.name, match.match) if isinstance(match, IgnoredMatch) else (match,)


def test_ignore_sha(cli_fs_runner):
    """
    GIVEN an empty cache and an empty config ignored_matches section
    WHEN I ignore a secret sha
    THEN the ignore secret is added to the config and saved
    """
    ignored_match = IgnoredMatch(
        name="test_name",
        match="41b8889e5e794b21cb1349d8eef1815960bf5257330fd40243a4895f26c2b5c8",
    )
    config = Config()

    with patch("ggshield.cmd.utils.context_obj.ContextObj.get") as mock_get_ctx:
        mock_get_ctx.return_value.config = config

        cmd = ["secret", "ignore", ignored_match.match, "--name", ignored_match.name]
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)

        assert_invoke_ok(result)
        assert config.user_config.secret.ignored_matches == [ignored_match]


def test_error_sha_last_found(cli_fs_runner):
    """
    GIVEN any config and cache
    WHEN I run the command with invalid arguments
    THEN an error should be raised
    """

    cmd = ["secret", "ignore", "some_secret_sha", "--last-found"]
    result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)

    assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)
    assert (
        "Option `--last-found` cannot be used with `SECRET_SHA` or `--name`."
        in result.output
    )


def test_error_ignore_sha_no_name(cli_fs_runner):
    """
    GIVEN any config and cache
    WHEN I run the command with a secret sha but no name
    THEN an error should be raised
    """

    cmd = ["secret", "ignore", "some_secret_sha"]
    result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)

    assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)
    assert "Option `--name` is required when ignoring a secret." in result.output


def test_cache_catches_last_found_secrets(client, isolated_fs):
    """
    GIVEN an empty cache and an empty config ignored_matches section
    WHEN I run a scan with multiple secrets
    THEN cache last_found_secrets is updated with these secrets and saved
    """
    commit = Commit.from_patch(_MULTIPLE_SECRETS_PATCH)
    config = Config()
    cache = Cache()
    cache.purge()
    assert cache.last_found_secrets == list()

    with my_vcr.use_cassette("multiple_secrets"):
        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=ScanContext(
                scan_mode=ScanMode.COMMIT_RANGE,
                command_path="external",
            ),
            secret_config=config.user_config.secret,
        )
        scanner.scan(commit.get_files(), scanner_ui=Mock())
    assert config.user_config.secret.ignored_matches == list()

    cache_found_secrets = sorted(cache.last_found_secrets, key=compare_matches_ignore)
    found_secrets = sorted(FOUND_SECRETS, key=compare_matches_ignore)

    assert [found_secret.match for found_secret in cache_found_secrets] == [
        found_secret.match for found_secret in found_secrets
    ]
    ignore_last_found(config, cache)
    for ignore in config.user_config.secret.ignored_matches:
        assert "test.txt" in ignore.name
    cache.load_cache()


def test_cache_catches_nothing(client, isolated_fs):
    """
    GIVEN a cache of last found secrets same as config ignored_matches
    WHEN I run a scan (therefore finding no secret)
    THEN config matches is unchanged and cache is empty
    """
    commit = Commit.from_patch(_MULTIPLE_SECRETS_PATCH)
    config = Config()
    config.user_config.secret.ignored_matches = FOUND_SECRETS
    cache = Cache()
    cache.last_found_secrets = FOUND_SECRETS

    with my_vcr.use_cassette("multiple_secrets"):
        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=ScanContext(
                scan_mode=ScanMode.COMMIT_RANGE,
                command_path="external",
            ),
            secret_config=config.user_config.secret,
        )
        results = scanner.scan(commit.get_files(), scanner_ui=Mock())

        assert sum(len(result.secrets) for result in results.results) == 0
        assert config.user_config.secret.ignored_matches == FOUND_SECRETS
        assert cache.last_found_secrets == []


def test_ignore_last_found(client, isolated_fs):
    """
    GIVEN a cache of last found secrets not empty
    WHEN I run a ignore last found command
    THEN config ignored_matches is updated accordingly
    """
    config = Config()

    cache = Cache()
    cache.last_found_secrets = FOUND_SECRETS
    ignore_last_found(config, cache)

    matches_ignore = sorted(
        config.user_config.secret.ignored_matches, key=compare_matches_ignore
    )

    found_secrets = sorted(FOUND_SECRETS, key=compare_matches_ignore)

    assert matches_ignore == found_secrets
    assert cache.last_found_secrets == FOUND_SECRETS


def test_ignore_last_found_with_manually_added_secrets(client, isolated_fs):
    """
    GIVEN a cache containing part of config ignored_matches secrets
    WHEN I run ignore command
    THEN only new discovered secrets are added to the config
    """
    manually_added_secret = (
        "41b8889e5e794b21cb1349d8eef1815960bf5257330fd40243a4895f26c2b5c8"
    )
    config = Config()
    config.user_config.secret.ignored_matches = [
        IgnoredMatch(name="", match=manually_added_secret)
    ]
    cache = Cache()
    cache.last_found_secrets = FOUND_SECRETS

    ignore_last_found(config, cache)

    matches_ignore = sorted(
        config.user_config.secret.ignored_matches, key=compare_matches_ignore
    )

    found_secrets = sorted(FOUND_SECRETS, key=compare_matches_ignore)
    assert matches_ignore == found_secrets


def test_do_not_duplicate_last_found_secrets(client, isolated_fs):
    """
    GIVEN 2 policy breaks on different files with the same ignore sha
    WHEN add_found_policy_break is called
    THEN only one element should be added
    """
    policy_break = PolicyBreak(
        "a", "Secrets detection", None, [Match("apikey", "apikey", 0, 0, 0, 0)]
    )
    cache = Cache()

    cache.add_found_policy_break(
        policy_break.break_type, get_ignore_sha(policy_break), "a"
    )
    cache.add_found_policy_break(
        policy_break.break_type, get_ignore_sha(policy_break), "b"
    )

    assert len(cache.last_found_secrets) == 1


def test_ignore_last_found_preserve_previous_config(client, isolated_fs):
    """
    GIVEN a cache containing new secrets AND a config not empty
    WHEN I run ignore command
    THEN existing config option are not wiped out
    """
    config = Config()
    previous_secrets = [
        IgnoredMatch(name="", match="previous_secret"),
        IgnoredMatch(name="", match="other_previous_secret"),
    ]

    previous_paths = {"some_path", "some_other_path"}
    config.user_config.secret.ignored_matches = previous_secrets.copy()
    config.user_config.secret.ignored_paths = previous_paths
    config.user_config.exit_zero = True

    cache = Cache()
    cache.last_found_secrets = FOUND_SECRETS
    ignore_last_found(config, cache)
    matches_ignore = sorted(
        config.user_config.secret.ignored_matches, key=compare_matches_ignore
    )

    found_secrets = sorted(FOUND_SECRETS + previous_secrets, key=compare_matches_ignore)

    assert matches_ignore == found_secrets
    assert config.user_config.secret.ignored_paths == previous_paths
    assert config.user_config.exit_zero is True



import concurrent.futures
import logging
import os
import sys
from ast import literal_eval
from concurrent.futures import Future
from typing import Dict, Iterable, List, Optional, Union

from pygitguardian import GGClient
from pygitguardian.models import APITokensResponse, Detail, MultiScanResult, TokenScope

from ggshield.core import ui
from ggshield.core.cache import Cache
from ggshield.core.client import check_client_api_key
from ggshield.core.config.user_config import SecretConfig
from ggshield.core.constants import MAX_WORKERS
from ggshield.core.errors import MissingScopesError, UnexpectedError, handle_api_error
from ggshield.core.scan import DecodeError, ScanContext, Scannable
from ggshield.core.scanner_ui.scanner_ui import ScannerUI
from ggshield.core.text_utils import pluralize

from .secret_scan_collection import Error, Result, Results


# GitGuardian API does not accept paths longer than this
_API_PATH_MAX_LENGTH = 256
_SIZE_METADATA_OVERHEAD = 10240  # 10 KB


logger = logging.getLogger(__name__)


if sys.version_info >= (3, 10):
    ScanFuture = Future[Union[Detail, MultiScanResult]]
else:
    ScanFuture = Future


class SecretScanner:
    """
    A SecretScanner scans a list of Scannable, using multiple threads
    """

    def __init__(
        self,
        client: GGClient,
        cache: Cache,
        scan_context: ScanContext,
        secret_config: SecretConfig,
        check_api_key: Optional[bool] = True,
    ):
        if check_api_key:
            check_client_api_key(client)

        self.client = client
        self.cache = cache
        self.secret_config = secret_config
        self.headers = scan_context.get_http_headers()
        self.headers.update({"scan_options": secret_config.dump_for_monitoring()})

        self.command_id = scan_context.command_id

        if secret_config.with_incident_details:
            response = self.client.api_tokens()

            if not isinstance(response, (Detail, APITokensResponse)):
                raise UnexpectedError("Unexpected api_tokens response")
            elif isinstance(response, Detail):
                raise UnexpectedError(response.detail)
            if TokenScope.INCIDENTS_READ not in response.scopes:
                raise MissingScopesError([TokenScope.INCIDENTS_READ])

    def scan(
        self,
        files: Iterable[Scannable],
        scanner_ui: ScannerUI,
        scan_threads: Optional[int] = None,
    ) -> Results:
        """
        Starts the scan, using at most `scan_threads`. If `scan_threads` is not set,
        defaults to MAX_WORKERS.
        Reports progress through `scanner_ui`.
        Returns a Results instance.
        """
        if scan_threads is None:
            scan_threads = MAX_WORKERS
        logger.debug("command_id=%s scan_threads=%d", self.command_id, scan_threads)

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=scan_threads, thread_name_prefix="content_scan"
        ) as executor:
            chunks_for_futures = self._start_scans(
                executor,
                files,
                scanner_ui,
            )

            return self._collect_results(scanner_ui, chunks_for_futures)

    def _scan_chunk(
        self, executor: concurrent.futures.ThreadPoolExecutor, chunk: List[Scannable]
    ) -> ScanFuture:
        """
        Sends a chunk of files to scan to the API
        """
        # `documents` is a version of `chunk` suitable for `GGClient.multi_content_scan()`
        documents = [
            {"document": x.content, "filename": x.filename[-_API_PATH_MAX_LENGTH:]}
            for x in chunk
        ]

        return executor.submit(
            self.client.multi_content_scan,
            documents,
            self.headers,
            all_secrets=True,
        )

    def _start_scans(
        self,
        executor: concurrent.futures.ThreadPoolExecutor,
        scannables: Iterable[Scannable],
        scanner_ui: ScannerUI,
    ) -> Dict[ScanFuture, List[Scannable]]:
        """
        Start all scans, return a tuple containing:
        - a mapping of future to the list of files it is scanning
        - a list of files which we did not send to scan because we could not decode them
        """
        chunks_for_futures = {}

        chunk: List[Scannable] = []
        max_payload_size = self.client.maximum_payload_size - _SIZE_METADATA_OVERHEAD
        utf8_encoded_chunk_size = 0
        maximum_document_size = int(
            os.getenv(
                "GG_MAX_DOC_SIZE",
                self.client.secret_scan_preferences.maximum_document_size,
            )
        )
        maximum_documents_per_scan = int(
            os.getenv(
                "GG_MAX_DOCS",
                self.client.secret_scan_preferences.maximum_documents_per_scan,
            )
        )
        logging.debug("max_doc_size=%d", maximum_document_size)
        logging.debug("max_docs=%d", maximum_documents_per_scan)
        for scannable in scannables:
            try:
                if scannable.is_longer_than(maximum_document_size):
                    scanner_ui.on_skipped(
                        scannable, f"content is over {maximum_document_size:,} bytes"
                    )
                    continue
                content = scannable.content
            except DecodeError:
                scanner_ui.on_skipped(scannable, "can't detect encoding")
                continue

            if content:
                if (
                    len(chunk) == maximum_documents_per_scan
                    or utf8_encoded_chunk_size + scannable.utf8_encoded_size
                    > max_payload_size
                ):
                    future = self._scan_chunk(executor, chunk)
                    chunks_for_futures[future] = chunk
                    chunk = []
                    utf8_encoded_chunk_size = 0
                chunk.append(scannable)
                utf8_encoded_chunk_size += scannable.utf8_encoded_size
            else:
                scanner_ui.on_skipped(scannable, "")
        if chunk:
            future = self._scan_chunk(executor, chunk)
            chunks_for_futures[future] = chunk
        return chunks_for_futures

    def _collect_results(
        self,
        scanner_ui: ScannerUI,
        chunks_for_futures: Dict[ScanFuture, List[Scannable]],
    ) -> Results:
        """
        Receive scans as they complete, report progress and collect them and return
        a Results.
        """
        self.cache.purge()

        results = []
        errors = []
        for future in concurrent.futures.as_completed(chunks_for_futures):
            chunk = chunks_for_futures[future]
            scanner_ui.on_scanned(chunk)

            exception = future.exception()
            if exception is None:
                scan = future.result()
            else:
                scan = Detail(detail=str(exception))
                errors.append(
                    Error(
                        files=[(x.filename, x.filemode) for x in chunk],
                        description=scan.detail,
                    )
                )

            if not scan.success:
                assert isinstance(scan, Detail)
                handle_scan_chunk_error(scan, chunk)
                continue

            assert isinstance(scan, MultiScanResult)
            for file, scan_result in zip(chunk, scan.scan_results):
                result = Result.from_scan_result(file, scan_result, self.secret_config)
                for secret in result.secrets:
                    self.cache.add_found_policy_break(
                        secret.detector,
                        secret.get_ignore_sha(),
                        file.filename,
                    )
                results.append(result)

        self.cache.save()
        return Results(results=results, errors=errors)


def handle_scan_chunk_error(detail: Detail, chunk: List[Scannable]) -> None:
    handle_api_error(detail)
    details = None

    ui.display_error("Scanning failed. Results may be incomplete.")
    try:
        # try to load as list of dicts to get per file details
        details = literal_eval(detail.detail)
    except Exception:
        pass

    if isinstance(details, list) and details:
        # if the details had per file details
        ui.display_error(
            f"Add the following {pluralize('file', len(details))}"
            " to your ignored_paths:"
        )
        for i, inner_detail in enumerate(details):
            if inner_detail:
                ui.display_error(f"- {chunk[i].filename}: {str(inner_detail)}")
        return
    else:
        # if the details had a request error
        filenames = "\n".join(f"- {file.filename}" for file in chunk)
        ui.display_error(f"The following chunk is affected:\n{filenames}")

        ui.display_error(str(detail))



import itertools
import logging
import traceback
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Callable, Iterable, Iterator, List, Pattern, Set

from click import UsageError
from pygitguardian import GGClient

from ggshield.core import ui
from ggshield.core.cache import Cache
from ggshield.core.client import check_client_api_key
from ggshield.core.config import Config
from ggshield.core.config.user_config import SecretConfig
from ggshield.core.constants import MAX_WORKERS
from ggshield.core.errors import ExitCode, QuotaLimitReachedError, handle_exception
from ggshield.core.scan import Commit, ScanContext
from ggshield.core.scanner_ui import create_message_only_scanner_ui
from ggshield.core.text_utils import STYLE, format_text
from ggshield.utils.git_shell import get_list_commit_SHA, is_git_dir
from ggshield.utils.os import cd

from .output import SecretOutputHandler
from .secret_scan_collection import Results, SecretScanCollection
from .secret_scanner import SecretScanner


logger = logging.getLogger(__name__)


# We add a maximal value to avoid silently consuming all threads on powerful machines
SCAN_THREADS = 4


def scan_repo_path(
    client: GGClient,
    cache: Cache,
    output_handler: SecretOutputHandler,
    exclusion_regexes: Set[Pattern[str]],
    config: Config,
    scan_context: ScanContext,
    repo_path: Path,
) -> int:  # pragma: no cover
    try:
        if not is_git_dir(repo_path):
            raise UsageError(f"{repo_path} is not a git repository")

        with cd(repo_path):
            return scan_commit_range(
                client=client,
                cache=cache,
                commit_list=get_list_commit_SHA("--all"),
                output_handler=output_handler,
                exclusion_regexes=exclusion_regexes,
                scan_context=scan_context,
                secret_config=config.user_config.secret,
            )
    except Exception as error:
        return handle_exception(error)


def scan_commits_content(
    commits: List[Commit],
    client: GGClient,
    cache: Cache,
    scan_context: ScanContext,
    secret_config: SecretConfig,
    progress_callback: Callable[[int], None],
    commit_scanned_callback: Callable[[Commit], None],
) -> SecretScanCollection:  # pragma: no cover
    try:
        commit_files = itertools.chain.from_iterable(c.get_files() for c in commits)

        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=scan_context,
            check_api_key=False,  # Key has been checked in `scan_commit_range()`
            secret_config=secret_config,
        )
        with create_message_only_scanner_ui() as scanner_ui:
            results = scanner.scan(
                commit_files, scan_threads=SCAN_THREADS, scanner_ui=scanner_ui
            )
    except QuotaLimitReachedError:
        raise
    except Exception as exc:
        logger.error(
            "Exception raised during scan. commits=%s type(exception)=%s exception=%s trace=%s",
            commits,
            type(exc),
            exc,
            traceback.format_exc(),
        )
        results = Results.from_exception(exc)
    finally:
        progress_callback(len(commits))
        for commit in commits:
            commit_scanned_callback(commit)

    result_for_urls = {result.url: result for result in results.results}
    scans = []
    for commit in commits:
        results_for_commit_files = [
            result_for_urls[u] for u in commit.urls if u in result_for_urls
        ]

        optional_header = (
            format_text(f"\ncommit {commit.sha}\n", STYLE["commit_info"])
            + f"Author: {commit.info.author} <{commit.info.email}>\n"
            + f"Date: {commit.info.date}\n"
        )

        scans.append(
            SecretScanCollection(
                commit.sha or "unknown",
                type="commit",
                results=Results(
                    results=results_for_commit_files,
                    errors=results.errors,
                ),
                optional_header=optional_header,
                extra_info={
                    "author": commit.info.author,
                    "email": commit.info.email,
                    "date": commit.info.date,
                },
            )
        )

    return SecretScanCollection(
        id=scan_context.command_id, type="commit-ranges", scans=scans
    )


def get_commits_by_batch(
    commits: Iterable[Commit],
    batch_max_size: int,
) -> Iterator[List[Commit]]:
    """
    Given a list of commit shas yield the commit files
    by biggest batches possible of length at most `batch_max_size`
    """
    current_count = 0
    batch = []
    for commit in commits:
        num_files = len(commit.urls)
        if current_count + num_files < batch_max_size:
            batch.append(commit)
            current_count += num_files
        else:
            # The first batch can remain empty if it has too many files
            if batch:
                yield batch
            current_count = num_files
            batch = [commit]
    # Send the last batch that remains
    yield batch


def scan_commit_range(
    client: GGClient,
    cache: Cache,
    commit_list: List[str],
    output_handler: SecretOutputHandler,
    exclusion_regexes: Set[Pattern[str]],
    scan_context: ScanContext,
    secret_config: SecretConfig,
    include_staged: bool = False,
) -> ExitCode:
    """
    Scan every commit in a range.

    :param client: Public Scanning API client
    :param commit_list: List of commits sha to scan
    :param verbose: Display successful scan's message
    """
    check_client_api_key(client)
    max_documents = client.secret_scan_preferences.maximum_documents_per_scan

    with ui.create_progress(len(commit_list) + int(include_staged)) as progress:
        commits_iters: list[Iterable[Commit]] = []
        if include_staged:
            commits_iters.append(
                (Commit.from_staged(exclusion_regexes=exclusion_regexes),)
            )
        commits_iters.append(
            (
                Commit.from_sha(sha, exclusion_regexes=exclusion_regexes)
                for sha in commit_list
            )
        )
        commits_batch = get_commits_by_batch(
            commits=itertools.chain(*commits_iters),
            batch_max_size=max_documents,
        )
        scans: List[SecretScanCollection] = []

        def commit_scanned_callback(commit: Commit):
            if include_staged and commit.sha is None:
                ui.display_verbose("Scanned staged changes")
            else:
                ui.display_verbose(f"Scanned {commit.sha}")

        with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
            futures = []
            for commits in commits_batch:
                futures.append(
                    executor.submit(
                        scan_commits_content,
                        commits,
                        client,
                        cache,
                        scan_context,
                        secret_config,
                        progress.advance,
                        commit_scanned_callback,
                    )
                )
                # Stop now if an exception has been raised by a future
                for future in futures:
                    exception = future.exception()
                    if exception is not None:
                        raise exception

            for future in as_completed(futures):
                scan_collection = future.result()
                for scan in scan_collection.scans_with_results:
                    if scan.results and scan.results.errors:
                        for error in scan.results.errors:
                            ui.display_error(error.description)
                    scans.append(scan)

    return_code = output_handler.process_scan(
        SecretScanCollection(
            id=scan_context.command_id, type="commit-range", scans=scans
        )
    )
    return return_code



from typing import List

import click


class NaturalOrderGroup(click.Group):
    """
    Helper class to force the order of the commands in the help text.
    Copied from https://github.com/pallets/click/issues/513#issuecomment-504158316
    This will not work with python < 3.6.
    """

    def list_commands(self, ctx: click.Context) -> List[str]:
        return list(self.commands.keys())



import platform
from collections import namedtuple
from unittest.mock import ANY, Mock, patch

import click
import pytest
from click import Command, Context, Group
from pygitguardian.models import (
    APITokensResponse,
    Detail,
    Match,
    MultiScanResult,
    PolicyBreak,
    ScanResult,
)

from ggshield import __version__
from ggshield.core.cache import Cache
from ggshield.core.config.user_config import SecretConfig
from ggshield.core.errors import (
    ExitCode,
    MissingScopesError,
    QuotaLimitReachedError,
    UnexpectedError,
)
from ggshield.core.scan import (
    Commit,
    DecodeError,
    ScanContext,
    ScanMode,
    Scannable,
    StringScannable,
)
from ggshield.core.scanner_ui.scanner_ui import ScannerUI
from ggshield.utils.git_shell import Filemode
from ggshield.utils.os import get_os_info
from ggshield.verticals.secret import SecretScanner
from ggshield.verticals.secret.secret_scanner import handle_scan_chunk_error
from tests.unit.conftest import (
    _MULTIPLE_SECRETS_PATCH,
    _NO_SECRET_PATCH,
    _ONE_LINE_AND_MULTILINE_PATCH,
    _SIMPLE_SECRET_TOKEN,
    API_TOKENS_RESPONSE_SCAN_SCOPES,
    GG_TEST_TOKEN,
    UNCHECKED_SECRET_PATCH,
    my_vcr,
)


ExpectedScan = namedtuple(
    "expectedScan", ("exit_code", "matches", "first_match", "want")
)
_EXPECT_NO_SECRET = {
    "content": "@@ -0,0 +1 @@\n+this is a patch without secret\n",
    "filename": "commit://patch/test",
    "filemode": Filemode.NEW,
}


@pytest.mark.parametrize(
    "name,input_patch,expected",
    [
        (
            "multiple_secrets",
            _MULTIPLE_SECRETS_PATCH,
            ExpectedScan(
                ExitCode.SCAN_FOUND_PROBLEMS, matches=4, first_match="", want=None
            ),
        ),
        (
            "simple_secret",
            UNCHECKED_SECRET_PATCH,
            ExpectedScan(
                ExitCode.SCAN_FOUND_PROBLEMS,
                matches=1,
                first_match=GG_TEST_TOKEN,
                want=None,
            ),
        ),
        (
            "one_line_and_multiline_patch",
            _ONE_LINE_AND_MULTILINE_PATCH,
            ExpectedScan(
                ExitCode.SCAN_FOUND_PROBLEMS, matches=1, first_match=None, want=None
            ),
        ),
        (
            "no_secret",
            _NO_SECRET_PATCH,
            ExpectedScan(
                exit_code=0, matches=0, first_match=None, want=_EXPECT_NO_SECRET
            ),
        ),
    ],
    ids=[
        "_MULTIPLE_SECRETS",
        "_SIMPLE_SECRET",
        "_ONE_LINE_AND_MULTILINE_PATCH",
        "_NO_SECRET",
    ],
)
def test_scan_patch(client, cache, name: str, input_patch: str, expected: ExpectedScan):
    commit = Commit.from_patch(input_patch)

    with my_vcr.use_cassette(name):
        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=ScanContext(
                scan_mode=ScanMode.PATH,
                command_path="external",
            ),
            secret_config=SecretConfig(),
        )
        results = scanner.scan(commit.get_files(), scanner_ui=Mock())
        for result in results.results:
            if result.secrets:
                assert len(result.secrets[0].matches) == expected.matches
                if expected.first_match:
                    assert result.secrets[0].matches[0].match == expected.first_match
            else:
                assert result.secrets == []

            if expected.want:
                assert result.filename == expected.want["filename"]
                assert result.filemode == expected.want["filemode"]


@pytest.mark.parametrize(
    "unscannable_type",
    [
        "EMPTY",
        "TOO_BIG",
        "BINARY",
    ],
)
def test_scanner_skips_unscannable_files(client, fs, cache, unscannable_type: str):
    """
    GIVEN a Scannable which is not scannable
    WHEN Scanner.scan() is called on it
    THEN it skips it
    AND the progress callback is called
    """
    mock = Mock(spec=Scannable)
    if unscannable_type == "EMPTY":
        mock.is_longer_than.return_value = False
        mock.content = ""
    elif unscannable_type == "TOO_BIG":
        mock.is_longer_than.return_value = True
    elif unscannable_type == "BINARY":
        mock.is_longer_than.side_effect = DecodeError

    scanner_ui = Mock(spec=ScannerUI)

    scanner = SecretScanner(
        client=client,
        cache=cache,
        scan_context=ScanContext(
            scan_mode=ScanMode.PATH,
            command_path="external",
        ),
        check_api_key=False,
        secret_config=SecretConfig(),
    )
    scanner.scan([mock], scanner_ui=scanner_ui)

    scanner_ui.on_skipped.assert_called_once()


def test_handle_scan_error_api_key():
    detail = Detail("Invalid API key.")
    detail.status_code = 401
    with pytest.raises(click.UsageError):
        handle_scan_chunk_error(detail, [])


@pytest.mark.parametrize(
    "detail, status_code, chunk",
    [
        pytest.param(
            Detail("Too many documents to scan"),
            400,
            [StringScannable(content="", url="/example") for _ in range(21)],
            id="too many documents",
        ),
        pytest.param(
            Detail(
                "[\"filename:: [ErrorDetail(string='Ensure this field has no more than 256 characters.', code='max_length')]\", '', '', '']"  # noqa
            ),
            400,
            [
                StringScannable(
                    content="still valid", url="/home/user/too/long/file/name"
                ),
                StringScannable(content="", url="valid"),
                StringScannable(content="", url="valid"),
                StringScannable(content="", url="valid"),
            ],
            id="single file exception",
        ),
    ],
)
def test_handle_scan_error(detail, status_code, chunk, capsys, snapshot):
    detail.status_code = 400
    handle_scan_chunk_error(detail, chunk)
    captured = capsys.readouterr()
    snapshot.assert_match(captured.err)


def test_handle_scan_quota_limit_reached():
    detail = Detail(detail="Quota limit reached.", status_code=403)
    with pytest.raises(QuotaLimitReachedError):
        handle_scan_chunk_error(detail, Mock())


def test_scan_merge_commit(client, cache):
    """
    GIVEN a merge commit in which a secret was inserted
    WHEN it is scanned
    THEN the secret is found
    """
    commit = Commit.from_patch(
        f"""
commit ca68e177596982fa38f181aa9944340a359748d2
Merge: 2c023f8 502d03c
Author: Aurelien Gateau <aurelien.gateau@gitguardian.com>
Date:   Thu Sep 5 18:39:58 2024 +0200

    Merge branch 'feature'
\0::100644 100644 100644 7601807 5716ca5 8c27e55 MM\0f\0\0diff --cc f
index 7601807,5716ca5..8c27e55
--- a/f
+++ b/f
@@@ -1,1 -1,1 +1,2 @@@
- baz
 -bar
++username=owly
++password={_SIMPLE_SECRET_TOKEN}
"""
    )

    with my_vcr.use_cassette("test_scan_merge_commit"):
        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=ScanContext(
                scan_mode=ScanMode.PATH,
                command_path="external",
            ),
            secret_config=SecretConfig(),
        )
        results = scanner.scan(commit.get_files(), scanner_ui=Mock())
        secrets = results.results[0].secrets
        assert len(secrets) == 1

        matches = {m.match_type: m.match for m in secrets[0].matches}
        assert matches["username"] == "owly"
        assert matches["password"] == _SIMPLE_SECRET_TOKEN


@pytest.mark.parametrize(
    "api_response, expected_exception, message",
    [
        (
            Detail(detail="Unexpected response"),
            UnexpectedError,
            "Unexpected response",
        ),
        (
            APITokensResponse.from_dict(API_TOKENS_RESPONSE_SCAN_SCOPES),
            MissingScopesError,
            "Token is missing the required scope incidents:read to perform this operation.",
        ),
    ],
)
def test_with_incident_details_error(
    monkeypatch, client, cache, api_response, expected_exception, message
):
    monkeypatch.setattr(client, "api_tokens", Mock(return_value=api_response))
    with pytest.raises(expected_exception) as exc_info:
        SecretScanner(
            client=client,
            cache=cache,
            scan_context=ScanContext(
                scan_mode=ScanMode.PATH,
                command_path="external",
            ),
            check_api_key=False,
            secret_config=SecretConfig(with_incident_details=True),
        )
        assert message in str(exc_info.value)


@patch("pygitguardian.GGClient.multi_content_scan")
def test_request_headers(scan_mock: Mock, client):
    """
    GIVEN a commit to scan
    WHEN SecretScanner.scan() is called on it
    THEN GGClient.multi_content_scan() is called with the correct values for
    `extra_headers`
    """
    c = Commit.from_patch(UNCHECKED_SECRET_PATCH)

    scan_result = ScanResult(policy_break_count=0, policy_breaks=[], policies=[])
    multi_scan_result = MultiScanResult([scan_result])
    multi_scan_result.status_code = 200
    scan_mock.return_value = multi_scan_result

    with Context(Command("bar"), info_name="bar") as ctx:
        os_name, os_version = get_os_info()
        ctx.parent = Context(Group("foo"), info_name="foo")
        scanner = SecretScanner(
            client=client,
            cache=Cache(),
            scan_context=ScanContext(
                scan_mode=ScanMode.PATH,
                command_path=ctx.command_path,
            ),
            check_api_key=False,
            secret_config=SecretConfig(),
        )
        scanner.scan(c.get_files(), scanner_ui=Mock())
    scan_mock.assert_called_with(
        ANY,
        {
            "GGShield-Version": __version__,
            "GGShield-Command-Path": "foo bar",
            "GGShield-Command-Id": ANY,
            "GGShield-OS-Name": os_name,
            "GGShield-OS-Version": os_version,
            "GGShield-Python-Version": platform.python_version(),
            "mode": "path",
            "scan_options": ANY,
        },
        all_secrets=True,
    )


@pytest.mark.parametrize("ignore_known_secrets", (True, False))
@patch("pygitguardian.GGClient.multi_content_scan")
def test_scan_ignore_known_secrets(scan_mock: Mock, client, ignore_known_secrets):
    """
    GIVEN a call multi_content_scan returning two policy breaks, one known and the other unknown
    WHEN -
    THEN the known policy break is ignored iff ignore_known_secrets is True
    """
    scannable = StringScannable(url="localhost", content="known\nunknown")
    known_secret = PolicyBreak(
        break_type="known",
        policy="Secrets detection",
        validity="valid",
        known_secret=True,
        matches=[
            Match(
                match="known",
                match_type="apikey",
                line_start=0,
                line_end=0,
                index_start=0,
                index_end=1,
            )
        ],
    )
    unknown_secret = PolicyBreak(
        break_type="unknown",
        policy="Secrets detection",
        validity="valid",
        known_secret=False,
        matches=[
            Match(
                match="unknown",
                match_type="apikey",
                line_start=0,
                line_end=0,
                index_start=0,
                index_end=1,
            )
        ],
    )

    scan_result = ScanResult(
        policy_break_count=1, policy_breaks=[known_secret, unknown_secret], policies=[]
    )
    multi_scan_result = MultiScanResult([scan_result])
    multi_scan_result.status_code = 200
    scan_mock.return_value = multi_scan_result

    scanner = SecretScanner(
        client=client,
        cache=Cache(),
        scan_context=ScanContext(
            scan_mode=ScanMode.PATH,
            command_path="ggshield",
        ),
        check_api_key=False,
        secret_config=SecretConfig(ignore_known_secrets=ignore_known_secrets),
    )
    results = scanner.scan([scannable], scanner_ui=Mock())

    if ignore_known_secrets:
        assert [pbreak.detector for pbreak in results.results[0].secrets] == ["unknown"]
    else:
        assert [pbreak.detector for pbreak in results.results[0].secrets] == [
            "known",
            "unknown",
        ]


@patch("pygitguardian.GGClient.multi_content_scan")
def test_scan_unexpected_error(scan_mock: Mock, client):
    """
    GIVEN a call multi_content_scan raising an exception
    WHEN calling scanner.scan
    THEN an UnexpectedError is raised
    """
    scannable = StringScannable(url="localhost", content="known\nunknown")

    scan_mock.side_effect = Exception("dummy")

    scanner = SecretScanner(
        client=client,
        cache=Cache(),
        scan_context=ScanContext(
            scan_mode=ScanMode.PATH,
            command_path="ggshield",
        ),
        check_api_key=False,
        secret_config=SecretConfig(),
    )
    with pytest.raises(UnexpectedError, match="Scanning failed.*"):
        scanner.scan([scannable], scanner_ui=Mock())


@patch("pygitguardian.GGClient.multi_content_scan")
def test_all_secrets_is_used(scan_mock: Mock, client):
    """
    GIVEN one secret ignored in backend, and one not ignored
    WHEN calling scanner.scan with the all_secrets option set to False
    THEN secrets excluded by the backend are ignored
    """
    scannable = StringScannable(url="localhost", content="known\nunknown")
    matches = [
        Match(
            match="known",
            match_type="apikey",
            line_start=0,
            line_end=0,
            index_start=0,
            index_end=1,
        )
    ]
    secret = PolicyBreak(
        break_type="not-excluded",
        policy="Secrets detection",
        validity="valid",
        matches=matches,
    )
    excluded_secret = PolicyBreak(
        break_type="excluded",
        policy="Secrets detection",
        validity="valid",
        matches=matches,
        is_excluded=True,
        exclude_reason="dummy",
    )

    scan_result = ScanResult(
        policy_break_count=1, policy_breaks=[secret, excluded_secret], policies=[]
    )
    multi_scan_result = MultiScanResult([scan_result])
    multi_scan_result.status_code = 200
    scan_mock.return_value = multi_scan_result

    scanner = SecretScanner(
        client=client,
        cache=Cache(),
        scan_context=ScanContext(
            scan_mode=ScanMode.PATH,
            command_path="ggshield",
        ),
        check_api_key=False,
        secret_config=SecretConfig(),
    )
    results = scanner.scan([scannable], scanner_ui=Mock())
    assert [pbreak.detector for pbreak in results.results[0].secrets] == [
        "not-excluded"
    ]



from collections.abc import Iterable

import pytest
from pygitguardian.models import ScanResult

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.filter import get_ignore_sha
from ggshield.core.scan import StringScannable
from ggshield.core.types import IgnoredMatch
from ggshield.verticals.secret import Results
from ggshield.verticals.secret.secret_scan_collection import (
    IgnoreKind,
    IgnoreReason,
    Result,
    compute_ignore_reason,
)
from tests.factories import PolicyBreakFactory, ScannableFactory, ScanResultFactory
from tests.unit.conftest import (
    _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
    _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_PATCH,
    _SIMPLE_SECRET_PATCH_SCAN_RESULT,
)


class MyException(Exception):
    pass


def test_results_from_exception():
    """
    GIVEN an exception
    WHEN creating a Results from it
    THEN it contains the right content
    """
    exc = MyException("Hello")
    results = Results.from_exception(exc)

    assert len(results.errors) == 1
    error = results.errors[0]
    assert error.description == "MyException: Hello"

    assert results.results == []


@pytest.mark.parametrize(
    ("content", "scan_result", "ignores", "final_len"),
    [
        pytest.param(
            _SIMPLE_SECRET_PATCH,
            _SIMPLE_SECRET_PATCH_SCAN_RESULT,
            [],
            _SIMPLE_SECRET_PATCH_SCAN_RESULT.policy_break_count,
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT-no remove, not all policies",
        ),
        pytest.param(
            _SIMPLE_SECRET_PATCH,
            _SIMPLE_SECRET_PATCH_SCAN_RESULT,
            ["2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93"],
            0,
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT-remove by sha",
        ),
        pytest.param(
            _SIMPLE_SECRET_PATCH,
            _SIMPLE_SECRET_PATCH_SCAN_RESULT,
            ["368ac3edf9e850d1c0ff9d6c526496f8237ddf91"],
            0,
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT-remove by plaintext",
        ),
        pytest.param(
            _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
            _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
            ["1945f4a0c42abb19c1a420ddd09b4b4681249a3057c427b95f794b18595e7ffa"],
            2,
            id="_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT-remove one by sha",
        ),
        pytest.param(
            _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
            _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
            [
                "060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e",
                "ce3f9f0362bbe5ab01dfc8ee565e4371",
            ],
            1,
            id="_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT-remove two by mix",
        ),
    ],
)
def test_create_result_removes_ignored_matches(
    content: str, scan_result: ScanResult, ignores: Iterable, final_len: int
) -> None:
    result = Result.from_scan_result(
        file=StringScannable(url="localhost", content=content),
        scan_result=scan_result,
        secret_config=SecretConfig(
            ignored_matches=[IgnoredMatch(name="", match=x) for x in ignores]
        ),
    )
    assert len(result.secrets) == final_len


@pytest.mark.parametrize("all_secrets", (True, False))
def test_create_result_removes_ignored_matches_bis(all_secrets):
    """
    GIVEN two different policy breaks
    WHEN ignoring the first one
    THEN it is ignored iff all_secrets is false

    Note: this test could replace the one above
    """
    scannable = ScannableFactory()
    policy_breaks = PolicyBreakFactory.create_batch(2, content=scannable.content)

    # ensure policy breaks are different
    if policy_breaks[0].matches[0].match_type == policy_breaks[1].matches[0].match_type:
        policy_breaks[0].matches[0].match_type += "a"

    config = SecretConfig(
        ignored_matches=[
            IgnoredMatch(name="x", match=get_ignore_sha(policy_breaks[0]))
        ],
        all_secrets=all_secrets,
    )
    result = Result.from_scan_result(
        scannable, ScanResultFactory(policy_breaks=policy_breaks), config
    )
    if all_secrets:
        assert len(result.secrets) == 2
        assert result.secrets[0].is_ignored is True
        assert result.secrets[1].is_ignored is False
    else:
        assert len(result.secrets) == 1
        assert result.secrets[0].is_ignored is False
        assert result.ignored_secrets_count_by_kind[IgnoreKind.IGNORED_MATCH] == 1


class TestComputeIgnoreReason:
    def test_ignore_excluded(self):
        """
        GIVEN an policy break excluded from the backend
        WHEN computing the ignore reason
        THEN it contains the original exclusion reason (and is not None)
        """
        policy_break = PolicyBreakFactory(
            is_excluded=True, exclude_reason="BACKEND_REASON"
        )
        assert compute_ignore_reason(policy_break, SecretConfig()) == IgnoreReason(
            IgnoreKind.BACKEND_EXCLUDED, "BACKEND_REASON"
        )

    def test_ignore_ignored_match(self):
        """
        GIVEN an policy break matching an ignored sha in config
        WHEN computing the ignore reason
        THEN it's not None
        """
        policy_break = PolicyBreakFactory()
        config = SecretConfig(
            ignored_matches=[
                IgnoredMatch(name="x", match=get_ignore_sha(policy_break))
            ],
        )
        assert compute_ignore_reason(policy_break, config) is not None

    def test_ignore_ignored_detector(self):
        """
        GIVEN a policy break matching an ignored detector in config
        WHEN computing the ignore reason
        THEN it's not None
        """
        policy_break = PolicyBreakFactory()
        config = SecretConfig(
            ignored_detectors=[policy_break.break_type],
        )
        assert compute_ignore_reason(policy_break, config) is not None

    @pytest.mark.parametrize("ignore_known", (True, False))
    def test_known_secret(self, ignore_known):
        """
        GIVEN a known policy break
        WHEN computing the ignore reason
        THEN it's not None iff ignore_secret is enabled in config
        """
        policy_break = PolicyBreakFactory(known_secret=True)
        config = SecretConfig(ignore_known_secrets=ignore_known)
        assert bool(compute_ignore_reason(policy_break, config)) is ignore_known



import click

from ggshield.verticals.hmsl.collection import NAMING_STRATEGIES, InputType


input_arg = click.argument(
    "path",
    type=click.Path(
        exists=True,
        dir_okay=False,
        readable=True,
        allow_dash=True,
    ),
)

naming_strategy_option = click.option(
    "--naming-strategy",
    "-n",
    type=click.Choice(list(NAMING_STRATEGIES.keys())),
    default="key",
    show_default=True,
    help="""Strategy to generate the hints in the output.

            \b
            - `censored`: only the first and last characters are displayed.
            - `cleartext`: the full secret is used as a hint (Not recommended!).
            - `none`: no hint is generated.
            - `key`: the key name is selected if available (e.g. in .env files), otherwise censored is used.""",
    callback=lambda _, __, value: NAMING_STRATEGIES[value],
)
input_type_option = click.option(
    "--type",
    "-t",
    "input_type",
    type=click.Choice(["file", "env"]),
    default="file",
    show_default=True,
    help="""Type of input to process.

            \b
            - `file`: the input is a simple file containing secrets.
            - `env`: the input is a file containing environment variables.""",
    callback=lambda _, __, value: InputType[value.upper()],
)
full_hashes_option = click.option(
    "-f",
    "--full-hashes",
    is_flag=True,
    default=False,
    help=(
        "Put the full hashes into the payload instead of the prefixes. This is useful"
        " for partners that trust GitGuardian because it allows to send more hashes"
        " per batch, and consumes less credits."
    ),
)



import json
import logging
from dataclasses import asdict
from typing import Any, List, TextIO, Tuple, cast

import click
from requests import HTTPError

from ggshield.cmd.hmsl.hmsl_common_options import input_arg
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.errors import ParseError, UnexpectedError
from ggshield.core.text_utils import pluralize
from ggshield.verticals.hmsl import HASH_REGEX, PREFIX_REGEX, get_client


logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@add_common_options()
@input_arg
def query_cmd(
    ctx: click.Context,
    path: str,
    **kwargs: Any,
) -> int:
    """
    Query HasMySecretLeaked using outputs from the `fingerprint` command.

    Note: If you used the `-f` option during stage one, the results will not be
    encrypted. It is still useful to pass them through `decrypt` to properly format the
    output.
    """

    # Opens the file or stdin
    input = cast(TextIO, click.open_file(path, "r"))
    payload, full_hashes = load_payload(input)

    # Get our client
    config = ContextObj.get(ctx).config
    client = get_client(config, ctx.command_path)

    # Send the hashes to the API
    try:
        for result in client.query(payload, full_hashes=full_hashes):
            line = json.dumps(asdict(result))
            click.echo(line)
    except (ValueError, HTTPError) as error:
        raise UnexpectedError(str(error))

    ui.display_info(
        f"{client.quota.remaining} {pluralize('credit', client.quota.remaining)} left for today."
    )
    ui.display_info(f"Queried {len(payload)} {pluralize('secret', len(payload))}.")
    return 0


def load_payload(input: TextIO) -> Tuple[List[str], bool]:
    """Load the payload from the input file.
    Return a list of string and a boolean indicating if they are full hashes (True)
    or prefixes (False).
    """
    # Remove empty lines
    data = list({line.strip() for line in input} - {""})
    # Check if we have full hashes or prefixes
    if not (
        all(HASH_REGEX.match(line) for line in data)
        or all(PREFIX_REGEX.match(line) for line in data)
    ):
        raise ParseError(
            "Invalid payload format. Did you forget to prepare your secrets?"
        )
    full_hashes = len(data) > 0 and len(data[0]) == 64
    return data, full_hashes



from pathlib import Path
from typing import List

from ggshield.core import ui
from ggshield.core.scan import Scannable


def print_file_list(files: List[Scannable], binary_paths: List[Path]) -> None:
    if not ui.is_verbose():
        return
    if binary_paths:
        ui.display_heading("Ignored binary files")
        for path in binary_paths:
            ui.display_verbose(f"- {path}")
    ui.display_heading("Files to scan")
    for f in files:
        ui.display_verbose(f"- {f.path}")
